"""Tia controllers

The TIA circuit on the board is a trans impedence amplifier.
https://en.wikipedia.org/wiki/Transimpedance_amplifier

It's a current to voltage converter and it's purpose is to convert current
to a signal our boards can measure (voltage).

Only a selected few boards have a TIA on-board.
To control the TIA use the get_tia_controller function.

"""
from .hdsoc import HdsocTIAController


def get_tia_controller(board, *args, **kwargs):
    """Get the correct TIA controller based on the board model.

    Currently only HDSoCv1 has a TIA controller.

    Args:
        board (Board): the board object.

    Returns:
        The appropriate TIA for the given board.
    """
    controller = {
        "hdsocv1": HdsocTIAController,
        "hdsocv1_evalr1": HdsocTIAController,
        "hdsocv1_evalr2": HdsocTIAController,
    }.get(board.model, None)
    if controller is None or not board.is_feature_enabled("tia_dac"):
        raise NotImplementedError(f"{board.model} does not support the TIAController")
    return controller(board, *args, **kwargs)



from .default import PedestalsGenerator
from .hdsoc import PedestalsGeneratorHdsoc
from .udc16 import PedestalsGeneratorUdc16
from .upac96 import PedestalsGeneratorUpac96


def get_pedestals_generator(
    board,
    num_captures: int = 10,
    num_warmup_events: int = 10,
    channels: "list[int]" = None,
):
    """Gets the appropriate pedestals generator for a board.

    Args:
        board (Board): the board
        num_captures (int): Number of datapoints per sample, used for averaging the values.
        num_warmup_events (int): Number of initial events to discard. Helps the board
            settle and excludes events that may skew the pedestals.
        channels (list[int]): channels to generate pedestals for. Defaults to all channels.

    Returns:
        The pedestals generator

    Raises:
        NotImplementedError if the given board does not support pedestals.
    """
    if not board.is_feature_enabled("pedestals"):
        raise NotImplementedError(
            f'Board "{board.model}" does not have support for pedestals.'
        )
    if board.model in ["upac32", "upaci", "zdigitizer"]:
        raise NotImplementedError("Board is currently unsupported")

    classy = {
        "hdsocv1_evalr2": PedestalsGeneratorHdsoc,
        "udc16": PedestalsGeneratorUdc16,
        "upac96": PedestalsGeneratorUpac96,
    }.get(board.model, PedestalsGenerator)
    return classy(board, num_captures, num_warmup_events, channels)



"""Pedestals generator for UPAC32.

UPAC32 works differently because the data cannot be captured in blocks.


NOTE: this does not work anymore, but it's left here for reference if UPAC32 ever rises from the dead
"""
import copy
import logging
from collections import deque

from naludaq.controllers import get_board_controller
from naludaq.tools import EventWaiter
from naludaq.tools.pedestals.pedestals_controller import (
    PedestalsController,
    _update_progress,
    sleep_calc,
)

LOGGER = logging.getLogger("naludaq.upac32_pedestal_generator")


class UpacPedestalsController(PedestalsController):
    """Pedestals controller for UPAC32"""

    def __init__(
        self,
        board,
        num_captures: int = 10,
        num_warmup_events: int = 10,
        channels: "list[int]" = None,
    ):
        super().__init__(board, num_captures, num_warmup_events)
        self.failed_validation = deque()

        self.margin = 2
        self.sleeptime = sleep_calc(self.board, self.margin)

    def _capture_data_for_pedestals(self, timeout: int = 3):
        """ """
        num_captures = self.num_captures
        num_warmup_events = self.num_warmup_events
        expected_amount = num_captures + num_warmup_events

        timeout_cnt = 0
        self.block_buffer = deque(maxlen=expected_amount)
        self._daq.output_buffer = self.block_buffer

        while len(self.validated_data) < num_captures and self._cancel is False:
            if timeout_cnt > timeout:
                break
            self.block_buffer = deque(maxlen=expected_amount)
            self._daq.output_buffer = self.block_buffer

            # Check the amount of data needed
            needed_amount = expected_amount - len(self.validated_data)
            # Extra code for status #############################################
            min_value = 0
            max_value = 80
            step_value = (max_value - min_value) / expected_amount
            #####################################################################
            # Capture data
            self._daq.start_capture()
            get_board_controller(self.board).start_readout()
            for idx in range(needed_amount + 1):
                if self._cancel:
                    break
                required_buffer_size = len(self._daq.output_buffer) + 1
                get_board_controller(self.board).toggle_trigger()

                _update_progress(
                    self.progress,
                    min_value + step_value * idx,
                    f"Capturing data, block: {len(self.block_buffer)+1}/{expected_amount}",
                )

                waiter = EventWaiter(
                    self._daq.output_buffer,
                    amount=required_buffer_size,
                    timeout=self.sleeptime,
                )
                try:
                    waiter.start(blocking=True)
                except TimeoutError:
                    LOGGER.warning("Event timed out")
            get_board_controller(self.board).stop_readout()
            self._daq.stop_capture()

            if len(self.block_buffer) != needed_amount:
                LOGGER.warning("data capture failed, increasing sleeptime")
                self._increase_sleeptime(needed_amount)

            # transfer data
            while self.block_buffer and len(self.validated_data) != num_captures:
                try:
                    event = (
                        self.block_buffer.pop()
                    )  # Pop from right skips warmup events
                except Exception as emsg:
                    continue
                LOGGER.debug("Transfer event")
                _update_progress(
                    self.progress,
                    min_value + step_value * len(self.validated_data),
                    f"Validate data {len(self.validated_data) + 1}/{expected_amount}",
                )

                # transfer validated data
                if event.get("data", None) is not None:
                    self.validated_data.append(event)

            if len(self.validated_data) == 0:
                LOGGER.debug(
                    "Pedestals buffer is empty, indicating transmission error, retrying."
                )
                timeout_cnt += 1
                continue
            elif self._store_warmup_events:
                self._warmup_data.append(self.block_buffer)

            # if len(self.validated_data) != needed_amount:
            #     LOGGER.debug("data capture failed, restarting")
            #     self._increase_sleeptime(needed_amount)

    def _backup_settings(self) -> dict:
        """Backup settings that might get overwritten to a dict.

        Returns:
            dict with the backup settings
        """
        backup = {
            "control_registers": copy.deepcopy(
                self.board.registers.get("control_registers", {})
            ),
        }

        return backup

    def _restore_backup_settings(self, backup_settings):
        """Restore all backuped settings to the board.

        Returns:
            True if settings have been restored, False if no old settings were found.
        """
        if not backup_settings:
            return

        for register in ["control_registers"]:
            reg = backup_settings.get(register, {})
            if reg:
                self.board.registers[register] = reg

    def _generate_pedestals_data(self, num_captures=None):
        """Generates the pedestals data from captured data.

        Uses captured events in the pedestals_raw_data deque
        to generate the data.

        Args:
            nCaptures (int): Number of captures, default 10
            blocksize (int): size of blocks, default 16
        """

        if num_captures is None:
            num_captures = self.num_captures

        block_size = self.board.params["pedestals_blocks"]

        self._reset_pedestals_data(num_captures)

        LOGGER.debug(
            "Generating pedestals from %s samples with blocksize: %s",
            num_captures,
            block_size,
        )

        return self._validate_and_transfer_data_to_pedestals(block_size, num_captures)

    def _validate_and_transfer_data_to_pedestals(self, block_size, num_captures):
        """Move data to board.pedestals after validating.

        Since data already is prevalidated this is just a logic check,

        Args:
            block_size(int): windows per block
            num_captures(int): amount of events averaged.
        """
        num_blocks = self._calculate_number_blocks(
            block_size, self.board.params["windows"]
        )
        for block in range(num_blocks):
            LOGGER.info("Block %s/%s", block, num_blocks)

            for cap in range(num_captures):
                try:
                    event = self.validated_data.popleft()
                except IndexError:
                    return False

                for chan in range(self.board.params["channels"]):
                    for window_num in range(block_size):
                        real_window = event["window_labels"][chan].index(window_num)
                        window = window_num + block * block_size
                        if block != 0 and (window < block_size):
                            continue
                        if window >= self.board.params["windows"]:
                            continue
                        for sample in range(self.board.params["samples"]):
                            index = sample + real_window * self.board.params["samples"]
                            data = event["data"][chan][index]
                            self.board.pedestals["rawdata"][chan][window][sample][
                                cap
                            ] = data

        return True

    def _store_time_metadata(self):
        """Store event times into pedestals metadata"""
        LOGGER.debug("Storing time metadata")
        self.metadata.store_event_times(self.validated_data)



import functools

try:
    import ftd2xx as ftd

    _FTDI_AVAILABLE = True
except:
    _FTDI_AVAILABLE = False

from naludaq.helpers.exceptions import FTDIError


def is_ftdi_available() -> bool:
    """Check if FTDI is available. This module can still be loaded without FTDI,
    so use this function if it is necessary to test the presence of drivers.
    """
    return _FTDI_AVAILABLE


def requires_ftd2xx(func):
    """Decorator. Use on functions to raise an FTDIError when FTDI is not present."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not _FTDI_AVAILABLE:
            raise FTDIError(
                "FTDI is not available because the drivers could not be found or loaded."
            )
        return func(*args, **kwargs)

    return wrapper


@requires_ftd2xx
def list_ftdi_devices(valid_only: bool = False, bytes_to_str: bool = False):
    """Lists all available FTDI devices.

    Args:
        valid_only (bool): remove devices that list as empty.
        bytes_to_str (bool): convert entries that are `bytes` to `str`.

    Returns:
        A dict of {dev_index: dev_detail_dict}.
    """
    if not isinstance(valid_only, bool):
        raise TypeError(
            f'Argument "valid_only" must be bool, not {type(valid_only).__name__}'
        )
    if not isinstance(bytes_to_str, bool):
        raise TypeError(
            f'Argument "bytes_to_str" must be bool, not {type(valid_only).__name__}'
        )

    devices = ftd.listDevices() or {}  # listDevices() can return None
    devices = {
        i: {
            k: v.decode() if bytes_to_str and type(v) is bytes else v
            for k, v in ftd.getDeviceInfoDetail(i).items()
        }
        for i, _ in enumerate(devices)
    }
    if valid_only:
        return {k: v for k, v in devices.items() if len(v.get("serial", "")) != 0}
    return devices


@requires_ftd2xx
def index_from_serial(value: str) -> int:
    """Get a device index given a part of a serial number. Case sensitive!

    Args:
        serial(str): Serial number to test, can be the shortened from the end.

    Returns:
        The device index that has a matching serial number.

    Raises:
        FTDIError if there was a problem listing devices or if there
        is no device with a matching serial number.
    """
    if not isinstance(value, str):
        raise TypeError("Serial Number must be a string.")

    try:
        return (
            [
                i
                for i, detail in list_ftdi_devices().items()
                if detail["serial"].decode()[-len(value) :] == value
            ]
        )[
            0
        ]  # Will raise IndexError if there is no match
    except ftd.DeviceError:
        raise FTDIError("Failed to scan FTDI devices.")
    except IndexError:
        raise FTDIError("No device exists with a matching serial number.")


@requires_ftd2xx
def index_from_comport(comport: "str | int") -> int:
    """Retrieves the device index from the com port it is connected to.

    Args:
        comport (str, int): the comport. If a string, must be formatted
            as either a '#' or 'com#', where "#" is the comport number.

    Returns:
        The device index

    Raises:
        ValueError if the given string is invalid
        TypeError if given an invalid argument type
        FTDIError if ftd2xx is unavailable or there is no device at the given com port.
    """
    if isinstance(comport, str):
        try:
            if "com" in comport.lower():
                comport = int(comport[3:])
            else:
                int(comport)
        except:
            raise ValueError("Invalid string supplied.")
    elif not isinstance(comport, int):
        raise TypeError("COM port must be a str or int.")

    ftdi_idx = -1
    for dev_idx in range(len(list_ftdi_devices())):
        try:
            ftdi_com = get_ftdi_com_port(dev_idx)
        except (FTDIError, ValueError):
            continue
        except NotImplementedError:
            break
        if ftdi_com == comport:
            ftdi_idx = dev_idx
            break
    if ftdi_idx == -1:
        raise FTDIError(f"No device at the given COM port: {comport}")
    return ftdi_idx


@requires_ftd2xx
def get_ftdi_com_port(device: "int | str") -> int:
    """Gets the COM port associated with an FTDI device

    Args:
        device (int, str): the device index or serial number

    Returns:
        The COM port as an int.

    Raises:
        TypeError if `device` is not an int or str.
        ValueError if there is no such device with the given serial number.
        FTDIError if there was a problem locating the device or opening it.
        NotImplementedError if the function isn't available for the system
    """
    if isinstance(device, int):
        index = device
    elif isinstance(device, str):
        try:
            index = index_from_serial(device)
        except (ValueError, FTDIError):
            raise
    else:
        raise TypeError("Device must be index (int) or serial number (str)")

    # ftd2xx exceptions are non-descriptive, need to split try-except blocks
    try:
        ser = ftd.open(index)  # not a context manager
    except ftd.DeviceError:
        raise FTDIError("Could not open the device at the given index.")

    try:
        comport = ser.getComPortNumber()
        if comport < 0:
            raise ValueError()
    except (ftd.DeviceError, ValueError):
        raise FTDIError("No COM port associated with index")
    except AttributeError:
        raise NotImplementedError("FT_GetComPortNumber is only available on windows")
    finally:
        ser.close()
    return comport


@requires_ftd2xx
def available_comports() -> list:
    """List available comports using the FTDI driver.

    Currently windows only.

    COM ports not using the FTDI driver will not show up.

    Returns:
        A list of comports

    Raises:
        FTDIError: if ftd2xx is unavailable
    """
    avail_ports = []
    # List available devices
    devices = ftd.listDevices() or {}
    for i, dev in enumerate(devices):
        # Open device
        try:
            ser = ftd.open(i)
        except ftd.DeviceError:
            pass
        else:
            # Read port num
            comnum = ser.getComPortNumber()

            comnum = f"{i}: COM{comnum}, {ser.getDeviceInfo()}"
            avail_ports.append(comnum)

            ser.close()

    return avail_ports



import logging

from .aodsoc import BoardControllerAodsoc

LOGGER = logging.getLogger("naludaq.board_controller_oleas")


class BoardControllerOleas(BoardControllerAodsoc):
    """Board controller for OLEAS (AODSOC AODS) board."""

    def set_oleas_enabled(self, en_trig: bool, en_a: bool, en_b: bool):
        """Set whether the oleas trigger or A/B are enabled."""
        self._write_control_register("oleas_en_trig", en_trig)
        self._write_control_register("oleas_en_a", en_a)
        self._write_control_register("oleas_en_b", en_b)

    def set_oleas_a(self, length: int, delay: int, polarity: int):
        """Sets the OLEAS A registers"""
        self._write_oleas_regs(0, length, delay, polarity)

    def set_oleas_b(self, length: int, delay: int, polarity: int):
        """Sets the OLEAS B registers"""
        self._write_oleas_regs(1, length, delay, polarity)

    def set_oleas_loop(self, loop: int):
        if not isinstance(loop, int):
            raise TypeError("Must be an int")
        if not 0 <= loop < 2**5:
            raise ValueError("Value out of bounds")
        self._write_control_register("oleas_loop", loop)

    def _write_oleas_regs(self, side: int, length: int, delay: int, polarity: int):
        """Write to the oleas A/B registers."""
        if not isinstance(length, int):
            raise TypeError("Length must be an int")
        if not isinstance(delay, int):
            raise TypeError("Length must be an int")
        if not isinstance(polarity, int):
            raise TypeError("Length must be an int or bool")
        if not 0 <= length < 2**16:
            raise ValueError("Length out of bounds")
        if not 0 <= delay < 2**16:
            raise ValueError("Delay out of bounds")
        if polarity not in [0, 1]:
            raise ValueError("Polarity must be 0 or 1")
        if side not in [0, 1]:
            raise ValueError("Side must be 0 or 1")
        suffix = {
            0: "a",
            1: "b",
        }[side]
        self._write_control_register(f"oleas_length_{suffix}", length)
        self._write_control_register(f"oleas_delay_{suffix}", delay)
        self._write_control_register(f"oleas_pol_{suffix}", polarity)



import logging

from naludaq.controllers.peripherals.peripherals_controller import PeripheralsController

logger = logging.getLogger("naludaq.peripherals.upac96")


class PeripheralsControllerUpac96(PeripheralsController):
    def __init__(self, board):
        """Peripherals controller for the UPAC96

        Contains a few functions to read from the two LTC2990 chips on the power regulator board.
        - read_udc1_1v2
        - read_udc1_2v5
        - read_udc2_1v2
        - read_udc2_2v5
        - read_fpga_1v0
        - read_fpga_1v8
        - read_fpga_io_2v5
        - read_fpga_io_3v3

        For reading temperatures:
        - read_readout_pcb_temperature
        - read_power_pcb_temperature
        """
        super().__init__(board)
        self._reader_functions = {
            "readout_pcb_temp": self.read_readout_pcb_temperature,
            "power_pcb_temp": self.read_power_pcb_temperature,
            "udc1_1v2": self.read_udc1_1v2,
            "udc1_2v5": self.read_udc1_2v5,
            "udc2_1v2": self.read_udc2_1v2,
            "udc2_2v5": self.read_udc2_2v5,
            "fpga_1v0": self.read_fpga_1v0,
            "fpga_1v8": self.read_fpga_1v8,
            "fpga_io_2v5": self.read_fpga_io_2v5,
            "fpga_io_3v3": self.read_fpga_io_3v3,
        }

    def read_readout_pcb_temperature(self):
        """Read the readout PCB temperature"""
        addr = (
            self.board.params["peripherals"]
            .get("readout_pcb_temp", {})
            .get("addr", 0b0011_000)
        )
        temp = self._read_temperature(addr)
        logger.debug("Temperature: %sC", temp)
        return temp

    def read_power_pcb_temperature(self):
        """Read the power PCB temperature"""
        addr = (
            self.board.params["peripherals"]
            .get("power_pcb_temp", {})
            .get("addr", 0b0011_001)
        )
        temp = self._read_temperature(addr)
        logger.debug("Temperature: %sC", temp)
        return temp

    def read_udc1_1v2(self):
        """Read V1P2_UDC1 (see schematic) from the board."""
        return self._read_ltc2990_voltage("udc1_1v2")

    def read_udc1_2v5(self):
        """Read V2P5_UDC1 (see schematic) from the board"""
        return self._read_ltc2990_voltage("udc1_2v5")

    def read_udc2_1v2(self):
        """Read V1P2_UDC2 (see schematic) from the board"""
        return self._read_ltc2990_voltage("udc2_1v2")

    def read_udc2_2v5(self):
        """Read V2P5_UDC2 (see schematic) from the board"""
        return self._read_ltc2990_voltage("udc2_2v5")

    def read_fpga_1v0(self):
        """Read V1P8_FPGA (see schematic) from the board"""
        return self._read_ltc2990_voltage("fpga_1v0")

    def read_fpga_1v8(self):
        """Read V1P8_FPGA (see schematic) from the board"""
        return self._read_ltc2990_voltage("fpga_1v8")

    def read_fpga_io_2v5(self):
        """Read V2P5_FPGA_IO (see schematic) from the board"""
        return self._read_ltc2990_voltage("fpga_io_2v5")

    def read_fpga_io_3v3(self):
        """Read V3P3_FPGA_IO (see schematic) from the board"""
        return self._read_ltc2990_voltage("fpga_io_3v3")



# Builtin
import gzip
import os
import pickle
from logging import getLogger
from pathlib import Path

# Internal
from naludaq.board import Board
from naludaq.tools.adc2mv.adc_linear_regression import ADCLinearReg
from naludaq.tools.dac_sweep.dac_sweep_controller import DACSweepController

# 3rd Party


LOGGER = getLogger(__name__)


def dac_sweep_and_lin_regression(
    board: Board,
    dac_step: int,
    filename: Path,
    minimum_r2: float = 0.9998,
    minimum_range: int = 10,
) -> dict:
    """
    Performs a dac sweep and calculates the linear region and
    linear regression. Stores the linear region and regression
    values to board.sample_lr and to file, and returns it.

    Args:
        board (naludaq.board): Board object to be used for dac sweep
        dac_step (int): Step size between dac values in dac counts
        filename (path): Location to store linear reg. values
        minimum_r2 (float): The minimum R-squared value for a region
            to be considered linear
        minimum_range (int): The minimum amount of points within a
            linear region
    Raises:
        TypeError if filename is None
        NotADirectoryError if the filename is not a valid path

    """
    if filename is None:
        raise TypeError("Supplied filename is NoneType.")
    path, _ = os.path.split(filename)
    if not os.access(os.path.dirname(filename), os.W_OK):
        raise NotADirectoryError(f"Not a valid file directory: {path}")

    dsc = DACSweepController(board)
    adclr = ADCLinearReg(board)
    output2analyze = dsc.dac_sweep(step_size=dac_step)
    lr_dict = adclr.linear_regression(output2analyze, minimum_r2, minimum_range)
    board.sample_lr = lr_dict

    # Save lr_list

    try:
        sfile = gzip.GzipFile(filename, "w", compresslevel=4)
        pickle.dump(lr_dict, sfile, protocol=pickle.HIGHEST_PROTOCOL)
    except IOError as error_msg:
        LOGGER.error("File could not be created. %s", error_msg)
        raise
    except pickle.PicklingError as error_msg:
        LOGGER.error("Saving lr_list failed: %s", error_msg)
        raise

    return lr_dict


def set_lr_dict(board: Board, filename: Path) -> bool:
    """
    Set the lr_dict for the board, to be used for adc2mv per sample.

    Args:
        board (naludaq.board): The board to load the lr_dict into
        filename (path): The location of the lr_dict file

    Returns:
        loaded (bool): True if file was successfully set, false otherwise.
    """
    loaded = True
    try:
        with gzip.open(filename, "rb") as fp:
            lr_list = pickle.load(fp)
            board.sample_lr = lr_list
    except IOError as error_msg:
        LOGGER.error("File could not be loaded. %s", error_msg)
        loaded = False
    except pickle.UnpicklingError as error_msg:
        LOGGER.error("Unpickling the data failed: %s", error_msg)
        loaded = False

    return loaded



import abc
from dataclasses import dataclass
from enum import Enum

from naludaq.backend.context import Context
from naludaq.backend.exceptions import DeviceError


class DeviceType(Enum):
    UDP = 0
    SERIAL = 1
    D2XX = 2
    D3XX = 3

    @staticmethod
    def from_string(type_name: str) -> "DeviceType":
        result = {
            "udp": DeviceType.UDP,
            "serial": DeviceType.SERIAL,
            "d2xx": DeviceType.D2XX,
            "d3xx": DeviceType.D3XX,
        }.get(type_name.lower(), None)
        if result is None:
            raise ValueError("Invalid device type")
        return result


class Device(abc.ABC):
    """Abstract wrapper for interacting with a device."""

    def __init__(self, context: Context, device_type: DeviceType):
        self._context = context
        self._type = device_type

    @property
    def context(self) -> Context:
        return self._context

    @property
    def is_open(self) -> bool:
        return self.is_valid

    @property
    def is_valid(self) -> bool:
        """Checks if this handle is usable"""
        return self._type == self._get_remote_type()

    @property
    def is_uart_based(self) -> bool:
        """Check if the connected device is UART-based (serial & D2XX).
        Will return `False` if there is no connection.
        """
        return hasattr(self, "baud_rate")

    @property
    def type(self) -> DeviceType:
        """Get the device type. This method does not check the backend."""
        return self._type

    @property
    def raw_connection_info(self) -> dict:
        """Get the raw connection info dict"""
        try:
            return self.raw_info["connection_info"]
        except KeyError:
            raise DeviceError("Device is not connected")

    @property
    def raw_info(self) -> dict:
        """Get the raw connection information dict"""
        return self._context.client.get_json("/connection/info")

    @property
    @abc.abstractmethod
    def info(self) -> dict:
        """Get connection info dict"""

    def clear_buffers(self):
        """Clear I/O buffers on the device."""
        self._context.client.put(
            "/connection/clear",
        )

    def _get_remote_type(self) -> DeviceType:
        """Get the type of device connected to the remote"""
        try:
            type_name = self.raw_info["connection_type"]
        except KeyError:
            raise DeviceError("Device is not connected")
        return DeviceType.from_string(type_name)

    def __repr__(self) -> str:
        """Generate string representation"""
        info = self.raw_connection_info
        info["valid"] = self.is_valid
        info = ", ".join(f"{k}={repr(v)}" for k, v in info.items())
        return f"{type(self).__name__}<{info}>"


class UdpDevice(Device):
    """Wrapper for working with a UDP device connected to the backend."""

    def __init__(self, context: Context):
        super().__init__(context, DeviceType.UDP)

    @property
    def board_ip(self) -> str:
        """Get the board IP address"""
        return self.raw_connection_info["board_ip"]

    @property
    def board_port(self) -> int:
        """Get the board port number"""
        return self.raw_connection_info["board_port"]

    @property
    def receiver_ip(self) -> str:
        """Get the receiver IP address"""
        return self.raw_connection_info["receiver_ip"]

    @property
    def receiver_port(self) -> int:
        """Get the receiver port number"""
        return self.raw_connection_info["receiver_port"]

    @property
    def info(self) -> dict:
        """Get connection info dict"""
        return {
            "type": "udp",
            "board_addr": (self.board_ip, self.board_port),
            "receiver_addr": (self.receiver_ip, self.receiver_port),
        }


class SerialDevice(Device):
    """Wrapper for working with a serial device connected to the backend."""

    def __init__(self, context: Context):
        super().__init__(context, DeviceType.SERIAL)

    @property
    def port(self) -> str:
        """Get the port the device is connected to"""
        return self.raw_connection_info["port"]

    @property
    def baud_rate(self) -> int:
        """Get/set the port baud rate"""
        return self.raw_connection_info["baud_rate"]

    @baud_rate.setter
    def baud_rate(self, value: int):
        self._context.client.put("/connection/serial", params={"baud_rate": value})
        if self.baud_rate != value:
            raise DeviceError("Failed to set baud rate")

    @property
    def rts_cts(self) -> bool:
        """Get/set whether RTS/CTS flow control is enabled."""
        return self.raw_connection_info["rts_cts"]

    @rts_cts.setter
    def rts_cts(self, enabled: bool):
        self._context.client.put("/connection/serial", params={"rts_cts": int(enabled)})
        if self.rts_cts != enabled:
            raise DeviceError("Failed to set flow control")

    @property
    def info(self) -> dict:
        """Get connection info dict"""
        return {
            "type": "serial",
            "port": self.port,
            "baud_rate": self.baud_rate,
        }


class D2xxDevice(Device):
    """Wrapper for working with a D2XX device connected to the backend."""

    def __init__(self, context: Context):
        super().__init__(context, DeviceType.D2XX)

    @property
    def serial_number(self) -> str:
        """Get the device serial number"""
        return self.raw_connection_info["serial_number"]

    @property
    def com_port(self) -> "str | None":
        """Get the device COM port.
        If there is no corresponding COM port (e.g. Linux), `None` is returned.
        """
        try:
            return self.raw_connection_info["com_port"]
        except KeyError:
            return None

    @property
    def baud_rate(self) -> int:
        """Get/set the device baud rate"""
        return self.raw_connection_info["baud_rate"]

    @baud_rate.setter
    def baud_rate(self, value: int):
        self._context.client.put("/connection/d2xx", params={"baud_rate": value})
        if self.baud_rate != value:
            raise DeviceError("Failed to set baud rate")

    @property
    def rts_cts(self) -> bool:
        """Get/set whether RTS/CTS flow control is enabled."""
        return self.raw_connection_info["rts_cts"]

    @rts_cts.setter
    def rts_cts(self, enabled: bool):
        self._context.client.put("/connection/d2xx", params={"rts_cts": int(enabled)})
        if self.rts_cts != enabled:
            raise DeviceError("Failed to set flow control")

    @property
    def info(self) -> dict:
        """Get connection info dict"""
        return {
            "type": "d2xx",
            "serial_number": self.serial_number,
            "baud_rate": self.baud_rate,
        }


class D3xxDevice(Device):
    """Wrapper for working with a D3XX device connected to the backend."""

    def __init__(self, context: Context):
        super().__init__(context, DeviceType.D3XX)

    @property
    def serial_number(self) -> str:
        """Get the device serial number"""
        return self.raw_connection_info["serial_number"]

    @property
    def description(self) -> str:
        """Get the device description"""
        return self.raw_connection_info["description"]

    @property
    def info(self) -> dict:
        """Get connection info dict"""
        return {
            "type": "d3xx",
            "serial_number": self.serial_number,
            "description": self.description,
        }


@dataclass
class AvailableSerialDevice:
    """Represents a serial device available to the backend"""

    port: str


@dataclass
class AvailableD2xxDevice:
    """Represents a D2XX device available to the backend"""

    serial_number: str
    com_port: "str | None"
    index: int
    description: str


@dataclass
class AvailableD3xxDevice:
    """Represents a D3XX device available to the backend"""

    serial_number: str
    description: str



"""Set board-level dac values.

The Nalu boards have DACs external to the chip. The DAC need to be set during boot.
Once the dacs have been set, it's easy to change them.

There are different types of dacs depending on the board model and revision.
This is abstracted away from the user by having a single user interface for all different boards.
By using a single interface the DACs can be set without the user needing to know details about
the board and the DAC chip specific implementations.

Note: Some boards do not have an onboard DAC chip.

The DACs in use are have different addressing and value space,
thus they need to be intialized and operated differently.

License: LGPL v3
See LICENSE.txt
"""
from naludaq.helpers.exceptions import InvalidBoardModelError

from .aardvarcv3 import DacControllerAardvarcv3
from .ad5671 import DACControllerAD5671
from .ad5675 import DACControllerAD5675
from .base import BaseDACController as _DACController
from .dac7578 import DACControllerDAC7578
from .hdsoc import DACControllerHDSoC
from .trbhm import DACControllerTRBHM
from .upac32 import DacControllerUpac32


def get_dac_controller(board) -> _DACController:
    """Gets the DAC controller appropriate for a given board.

    Args:
        board (Board): the board object.

    Returns:
        The DAC controller

    Raises:
        InvalidBoardModelError if the given board does not have
            a DAC controller
    """
    if not board.is_feature_enabled("ext_dac"):
        raise InvalidBoardModelError("The given board does not support DAC control")

    controller = {
        "aardvarcv3": DacControllerAardvarcv3,
        "aardvarcv4": DacControllerAardvarcv3,
        "aodsv1": DACControllerAD5671,
        "aodsv2_eval": DACControllerAD5671,
        "asocv3": DACControllerAD5671,
        "asocv3s": DACControllerAD5671,
        "hdsocv1": DACControllerHDSoC,
        "hdsocv1_evalr1": DACControllerHDSoC,
        "hdsocv1_evalr2": DACControllerHDSoC,
        "trbhm": DACControllerTRBHM,
        "udc16": DACControllerDAC7578,
        "upac32": DacControllerUpac32,
        "upac96": DACControllerDAC7578,
        "aodsoc_aods": DACControllerDAC7578,
        "aodsoc_asoc": DACControllerDAC7578,
    }.get(board.model, None)
    if controller is None:
        raise InvalidBoardModelError(f"No DAC controller for {board.model}")

    return controller(board)



def get_data_collector(board):
    if board.model in ["upac96", "udc16"]:
        from .udc16 import Udc16DataCollector

        return Udc16DataCollector(board)
    else:
        from .default import DataCollector

        return DataCollector(board)



"""
"""
from logging import getLogger

from naludaq.communication import ControlRegisters, DigitalRegisters

from .default import BoardController

LOGGER = getLogger("naludaq.board_controller_trbhm")


class TrbhmBoardController(BoardController):
    """Special board controller for TRBHM."""

    def read_scalar(self, channel: int) -> int:
        """Read the scalar for the given channel"""
        channels_per_chip = self.board.channels // self.board.available_chips
        relative_channel = channel % channels_per_chip
        chip = channel // channels_per_chip
        return self._read_scalar_inner(chip, relative_channel)

    def _read_scalar_inner(self, chip: int, relative_channel: int) -> int:
        """Read the scalar for the given channel.

        This is the inner function that actually reads the scalar.
        It is called by `read_scalar` and should not be called directly.
        """
        name = self.get_scal_name(relative_channel)
        scal = self._read_digital_register(name, chips=chip)
        try:
            scalhigh = self._read_digital_register("scalhigh", chips=chip)
        except (KeyError, AttributeError):
            scalhigh = 0
        shift_amt = DigitalRegisters(self.board).registers[name]["bitwidth"]
        scal += scalhigh << shift_amt

        return scal

    def toggle_trigger(self):
        """Toggles the ext trigger using software.

        For TRBHM the wait between separate register writes is too long, and
        toggling the trigger too slowly results in too many events coming back
        and filling the FIFO, causing malformed events. This method instead
        sends the register writes all as one string.
        """
        cr = ControlRegisters(self.board)

        wait_cmd = "AE000001"
        exttrig_high_cmd = cr.generate_write("exttrig", True)
        exttrig_low_cmd = cr.generate_write("exttrig", False)
        toggle_cmd = wait_cmd + exttrig_high_cmd + exttrig_low_cmd
        self._send_command(toggle_cmd)

    def set_loopback_enabled(self, enabled: bool):
        """Set whether serial loopback is enabled.

        Loopback can safely be disabled during most of the operations with the board.
        Loopback **must** be disabled when communicating over the serial interface.
        If serial communication with the ASIC is intended then this should run during startup and only be enabled as needed.

        Args:
            enabled (bool): True to enable loopback.

        Raises:
            TypeError if enabled is not a bool.
        """
        if not isinstance(enabled, bool):
            raise TypeError("Argument must be bool")
        OFF = "B0900002"
        ON = "B0900003"
        cmd = ON if enabled else OFF
        self._send_command(cmd)



import logging
import time

import numpy as np

from naludaq.communication import control_registers, digital_registers
from naludaq.controllers import get_dac_controller
from naludaq.helpers.exceptions import (
    BoardParameterError,
    OperationCanceledError,
    PedestalsDataCaptureError,
)
from naludaq.tools.pedestals import get_pedestals_controller

LOGGER = logging.getLogger(__name__)  # pylint: disable=invalid-name
DigitalRegisters = digital_registers.DigitalRegisters
ControlRegisters = control_registers.ControlRegisters


class DACSweepController:
    """DAC Sweep Controller allows the user to easily collect data for
    a sweep of dac values.

    The DAC Sweep Controller uses portions of the pedestal controller
    to collect data for every sample per channel per dac sweep, allowing
    the user to find the linear region and linear regression values on
    a per sample basis.

    Args:
        board (naludaq.board): Board object

    Attributes:
        num_attempts: How many attempts to try a DAC value before skipping that dac value

    raises:
        BoardParameterError if ext_dac fields are not available amoung the board params.
    """

    def __init__(self, board):
        # The type of daq used depends on the connection type
        self._board = None
        self.board = board
        self._num_attempts = 2
        self._dac_min_counts: int = 0
        self._dac_max_counts: int = self._get_max_counts()
        self._progress: list = []
        self._cancel = False
        self.ped_ctrl = get_pedestals_controller(board)
        self._backup_dac: dict = {}

    def _get_max_counts(self) -> int:
        try:
            return self.board.params["ext_dac"]["max_counts"]
        except KeyError as e:
            raise BoardParameterError(
                f"The register file lacks necessary ext_dac fields: {e}"
            )

    @property
    def dac_max_counts(self):
        """Get/Set the max count value for the DAC.

        The true max value is determined by the board params.

        12-bit: <= 4095
        16-bit: <= 65535

        Raises:
            TypeError: if value is not an int
            ValueError: if value is outside of the boards accepted range.
        """
        return self._dac_max_counts

    @dac_max_counts.setter
    def dac_max_counts(self, value):
        if not isinstance(value, int):
            raise TypeError("dac max counts must be an int.")
        max_val = self._get_max_counts()
        if not self.dac_min_counts <= value <= max_val:
            ValueError(
                f"dac_max_counts must be a value between {self.dac_min_counts} and {max_val}"
            )
        self._dac_max_counts = value

    @property
    def dac_min_counts(self):
        """Get/Set the min count value for the DAC.

        The min count for a board should be 0, unless otherwise stated

        Raises:
            TypeError: if value is not an int
            ValueError: if value is outside of the boards accepted range.
        """
        return self._dac_min_counts

    @dac_min_counts.setter
    def dac_min_counts(self, value):
        if not isinstance(value, int):
            raise TypeError("dac min counts must be an int.")
        min_val = self.board.params["ext_dac"].get("min_counts", 0)
        if not min_val <= value <= self.dac_max_counts:
            ValueError(
                f"dac_min_counts must be a value between {min_val} and {self.dac_max_counts}"
            )
        self._dac_min_counts = value

    @property
    def num_attempts(self):
        return self._num_attempts

    @num_attempts.setter
    def num_attempts(self, value):
        if not isinstance(value, int):
            raise TypeError("num_attempts must be an int.")
        if value < 0:
            raise ValueError("num_attempts must be positive")
        self._num_attempts = value

    @property
    def board(self):
        return self._board

    @board.setter
    def board(self, board):
        self._board = board

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError("Progress is stored in a list")
        self._progress = value

    def cancel(self):
        """Cancels the dac sweep as soon as possible.
        No data is generated, and the board is restored to
        its previous state.

        Can only be called from a separate thread.
        """
        self._cancel = True
        if self.ped_ctrl:
            self.ped_ctrl.cancel()

    def dac_sweep(
        self,
        step_size: int = 50,
        min_counts: int = 0,
        max_counts: int = None,
        num_captures: int = 3,
        num_warmup_events: int = 1,
        channels: list = None,
    ):
        """Perform a dac sweep with a given step size.
        IMPORTANT: TURN OFF CONNECTED FUNCTION GENERATORS BEFORE RUNNING
        Ext-dac + signal = :(

            Args:
                step_size (int): Step sizes for the ext-dac, from 0 to max_counts
                min_counts (int): minimum DAC value to start sweep from
                max_counts (int): maximum DAC value to stop sweep at
                num_captures (int): number of events per data point
                num_warmup_events (int): number of events to capture before starting the sweep
                channels (list): list of channel numbers to sweep over

            Returns:
                dac_sweep (dict): In form of:
                    dac_val (int): pedestals (dict) {
                        "data", "rawdata"
                    }
                or None if the dac sweep was canceled.
        """
        self._cancel = False

        if step_size <= 0 or step_size > self.dac_max_counts:
            raise ValueError(f"Step size is out of bounds: 0-{self.dac_max_counts}")
        if channels is None:
            channels = list(range(self.board.channels))
        if max_counts is None:
            max_counts = self.board.params["ext_dac"]["max_counts"]

        self._backup_settings()

        output = {}

        ext_dac_values = np.arange(min_counts, max_counts + 1, step_size)
        LOGGER.info(
            f"Starting DAC Sweep from {min_counts}-{max_counts} with step {step_size}"
        )

        for dac_idx, dac_val in enumerate(ext_dac_values):
            dac_val = int(dac_val)
            LOGGER.info(f"Currently running DAC_val: {dac_val}")
            # set dac value for all channels
            self.progress.append(
                (
                    int(98 * (dac_val - min_counts) / (max_counts - min_counts)),
                    f"Collecting DAC value {dac_idx+1}/{len(ext_dac_values)}",
                )
            )
            dac_data = self._run_dac_value(
                dac_val, channels, num_captures, num_warmup_events
            )
            if dac_data:
                output[dac_val] = dac_data
            if self._cancel:
                break

        # Restore backups
        self._restore_settings()

        # Data is probably bad if canceled, so return None
        if self._cancel:
            return None

        # save data
        try:
            self.progress.append((99, "Saving DAC Sweep"))
        except:
            LOGGER.debug("Status message couldn't append to self.progress")
        return output

    def _run_dac_value(self, dac_val, channels, num_captures, num_warmup_events):
        """Run a single dac value for all channels.

        Args:
            dac_val (int): DAC value to set
            channels (list): list of channel numbers to sweep over
            num_captures (int): number of events per data point
            num_warmup_events (int): number of events to capture before starting the sweep

        Returns:
            output (dict): In form of:
                channel (int): {
                    "data": (np.array) pedestal data,
                    "rawdata": (np.array) raw pedestal data
                }
        """
        self._set_dacs(dac_val, channels)
        output = None
        for _ in range(self.num_attempts):
            try:
                output = self._capture_pedestals(num_captures, num_warmup_events)
                LOGGER.info(f"{dac_val} total data: {len(output)}")
                break
            except PedestalsDataCaptureError:
                LOGGER.error("Couldn't capture %d DAC data", dac_val)
                continue
            except (KeyboardInterrupt, OperationCanceledError):
                self.cancel()
                break
        return output

    def _set_dacs(self, dac_val, channels):
        get_dac_controller(self.board).set_dacs(dac_val, channels)
        time.sleep(0.01)

    def _capture_pedestals(self, num_captures: int, num_warmup_events: int):
        """Capture pedestals data for each data point in the DAC sweep.

        This function will behave a bit different if it's a

        Args:
            num_captures (int): number of events per data point
            num_warmup_events (int): number of events to capture before starting the sweep

        Returns:
            pedestals (dict): In form of:
                "data": {
                    channel (int): {
                        "data": (np.array) pedestal data,
                        "rawdata": (np.array) raw pedestal data
                    }
                }
        """
        self.ped_ctrl.reset_pedestals()
        self.ped_ctrl.num_captures = num_captures
        self.ped_ctrl.num_warmup_events = num_warmup_events
        blocks = self.ped_ctrl._capture_data_for_pedestals()
        self.ped_ctrl._store_raw_data(blocks)
        self.ped_ctrl._store_averaged_data()
        return self.board.pedestals

    def _backup_settings(self):
        self._backup_pedestals = self.board.pedestals
        self._backup_dac = (
            self.board.params.get("ext_dac", {}).get("channels", {}).copy()
        )

    def _restore_settings(self):
        self.board.pedestals = self._backup_pedestals
        for chan, value in self._backup_dac.items():
            get_dac_controller(self.board).set_single_dac(chan, value)



"""
"""


from naludaq.controllers.trigger.default import TriggerController


class TriggerControllerSiread(TriggerController):
    def __init__(self, board):
        super().__init__(board)
        self.banks = 4



from . import answer_parser
from .aardvarcv3_parser import Aardvarcv3Parser
from .aodsoc_parser import OddsockParser
from .asocv3_parser import ASoCv3Parser
from .asocv3s_parser import ASoCv3SParser
from .hdsoc_parser import HDSoCParser
from .headers import get_header_parser
from .parser import Parser
from .trbhm_parser import TRBHMParser
from .udc_parser import UDCParser
from .upac96_parser import Upac96Parser
from .upac_parser import UPACParser

PARSERS = {
    "aardvarcv3": Aardvarcv3Parser,
    "aardvarcv4": Aardvarcv3Parser,
    "aodsoc_aods": OddsockParser,
    "aodsoc_asoc": OddsockParser,
    "asocv3": ASoCv3Parser,
    "asocv3s": ASoCv3SParser,
    "hdsocv1": HDSoCParser,
    "hdsocv1_evalr1": HDSoCParser,
    "hdsocv1_evalr2": HDSoCParser,
    "trbhm": TRBHMParser,
    "udc16": UDCParser,
    "upac32": UPACParser,
    "upac96": Upac96Parser,
    "upaci": UPACParser,
    "zdigitizer": UPACParser,
}


def get_parser(params):
    """Return a parser object for the board.

    Args:
        params (dict): Dictionary with the parser parameters.
        These are generally part of the board.params.

    Returns:
        instantiated parser object. ready to roll!
    """
    model = params.get("model", "default")
    p = PARSERS.get(model, Parser)

    return p(params)



"""
"""
def get_trigger_controller(board):
    """Get the controller for the model you are using.

    This is a small factory to give you the correct class without the user caring.

    Args:
        board (Board): the board object.

    Returns:
        Instantiated TriggerController
    """
    if board.model in ["aodsoc_aods", "aodsoc_asoc"]:
        from naludaq.controllers.trigger.aodsoc import TriggerControllerAodsoc
        return TriggerControllerAodsoc(board)
    elif board.model in ["hdsocv1", "hdsocv1_evalr1", "hdsocv1_evalr2"]:
        from naludaq.controllers.trigger.hdsoc import TriggerControllerHdsoc
        return TriggerControllerHdsoc(board)
    elif board.model in ["upac32", "upaci", "zdigitizer"]:
        from naludaq.controllers.trigger.upac import TriggerControllerUpac
        return TriggerControllerUpac(board)
    elif board.model == "upac96":
        from naludaq.controllers.trigger.upac96 import TriggerControllerUpac96
        return TriggerControllerUpac96(board)
    elif board.model == "siread":
        from naludaq.controllers.trigger.siread import TriggerControllerSiread
        return TriggerControllerSiread(board)
    elif board.model == "trbhm":
        from naludaq.controllers.trigger.trbhm import TriggerControllerTrbhm
        return TriggerControllerTrbhm(board)
    else:
        from naludaq.controllers.trigger.default import TriggerController
        return TriggerController(board)



"""
"""
import time
from logging import getLogger
from typing import Iterable

from naludaq.backend.managers.connection import ConnectionManager
from naludaq.backend.managers.io import BoardIoManager
from naludaq.backend.models.device import DeviceType
from naludaq.communication import ControlRegisters, DigitalRegisters
from naludaq.controllers.controller import Controller
from naludaq.helpers import type_name

LOGGER = getLogger("naludaq.board_controller_default")
NON_UDP = [DeviceType.SERIAL, DeviceType.D2XX, DeviceType.D3XX]


class BoardController(Controller):
    """The board control communicates with to the board.

    This class is stateless, you need to supply the board to the class to perform actions on it.
    There is no need to store the controller, it will only be used to control the board.

    It is the interface between the application and the board.
    The Analytics package or the GUI uses this class to communicate with the
    hardware. Keeps track of all registers etc.
    It doesn't receive continous data, the data acquisition module should be used for that.
    """

    def __init__(self, board):
        super().__init__(board)
        self._trigger_wait_cycles = board.params.get("ext_trig_cycles", 3)

    def start_readout(
        self,
        trig="self",
        lb="trigrel",
        acq="raw",
        dig_head=False,
        ped="zero",
        readoutEn=True,
        singleEv=False,
    ):
        """Sends a readout signal to the board.

        Obstensibly reads out the chip?

        Args:
            trigger_type (str):
                'immediate', '00' - digitize whatever instantly as fast as possible.
                'ext': '01' - use the external trigger.
                'self': '10' - trigger on input RF signal.
            lbType (str):
                'forced': '00' - always reads out same set of windows, start window set by lookback
                'trig': '01' - relative to the trigger.
                'roi': '10' - complicated...
            acq_type (str):
                'raw': '00' - Can't ped-subtract. <Default output>
                'ped': '01' - sub, spits out headers, channels, and windows. <Requires firmware modification>
            dig_head (bool):
                True: reads out a "digitization" header after each sram read (only works with acq_type="ped")
                False: doesn't read out that header
            pedMode (str):
                'z': '00' - all peds = 0
                'c': '01' - peds = chan*1024, set offset per channel
                'r': '11' - peds = RAM addr
            readoutEn (bool): Enables readout to ethernet (otherwise data thrown away)
                self.readReadoutRate() is the companion to this to see the true rate
                readout rate scalar stored in FPGA reg N_GPR(32) + 5
            singleEV (bool): True - only reads one event.

        Command
        -------

        command is: BMMXXVVV
        MM = singleEvent(7) + readoutEn(6) + pedSel(5 dt 4) + modeSel(3 dt 0)
        - modeSel=x"0" analog write
        - modeSel=x"1" digital write
        - modeSel=x"2" digital read
        - modeSel=x"3" waveform read

        VVVVV = addr_sel(19 dt 12) + data_sel(11 dt 0)
        - data_sel(5 downto 4) = trig type
        - data_sel(3 downto 2) = lb type
        - data_sel(1 downto 0) = acq type

        """
        readout_params = self._validate_readout_params(
            trig, lb, acq, dig_head, ped, readoutEn, singleEv
        )
        cmd = self._generate_start_readout_cmd(readout_params)
        LOGGER.debug("READOUT command: %s", cmd)

        self._send_command(cmd)

    @staticmethod
    def _validate_readout_params(
        trig, lb, acq, dig_head, ped, readoutEn, singleEv
    ) -> dict:
        for name, in_val in {"trig": trig, "lb": lb, "acq": acq, "ped": ped}.items():
            if not isinstance(in_val, str):
                raise TypeError(f"{name} is {type(in_val)}, expected str")
        if not trig[0] in ["i", "e", "s"]:
            raise ValueError(
                f"trig not valid, got {trig}, "
                "options are 'immediate', external', 'self'"
            )
        if not lb[0] in ["f", "t", "r"]:
            raise ValueError(
                f"Invalid lbType, got {lb}, "
                "options are 'forced', 'trigrel', or 'roi'"
            )
        if not acq[0] in ["r", "p"]:
            raise ValueError(
                f"invalid acq_type, got {acq}: " "options are 'raw', or 'pedsub'"
            )
        if not ped[0] in ["z", "c", "r"]:
            raise ValueError(
                f"invalid pedestals_type, got {ped}: "
                "options are 'zero', 'chanoffset', or 'ramaddr'"
            )

        if not isinstance(readoutEn, bool):
            raise TypeError(f"readoutEn is {type(readoutEn)}, expected bool")
        if not isinstance(singleEv, bool):
            raise TypeError(f"singleEv is {type(singleEv)}, expected bool")
        if not isinstance(dig_head, bool):
            raise TypeError(f"dig_head is {type(dig_head)}, expected bool")

        # check they are all there
        # check validity
        readout_params = dict()
        readout_params["trig"] = trig
        readout_params["lb"] = lb
        readout_params["acq"] = acq
        readout_params["ped"] = ped
        readout_params["readoutEn"] = readoutEn
        readout_params["singleEv"] = singleEv
        readout_params["dig_head"] = dig_head

        return readout_params

    def _generate_start_readout_cmd(self, readout_params):

        (
            trigger_type,
            lbType,
            acq_type,
            pedestals_type,
        ) = self._convert_readout_params_to_binary(readout_params)

        readoutEn = readout_params.get("readoutEn", None)
        singleEv = readout_params.get("singleEv", None)
        dig_head = readout_params.get("dig_head", None)

        first_part = hex(
            int(str(int(singleEv)) + str(int(readoutEn)) + pedestals_type + "0011", 2)
        )[2:].zfill(2)

        second_part = hex(
            int("000000" + trigger_type + lbType + str(int(dig_head)) + acq_type, 2)
        )[2:].zfill(3)

        return f"B{first_part}00{second_part}"

    @staticmethod
    def _convert_readout_params_to_binary(readout_params):
        trig = readout_params.get("trig", None)
        lb = readout_params.get("lb", None)
        acq = readout_params.get("acq", None)
        ped = readout_params.get("ped", None)

        trigger_type = lbType = acq_type = pedestals_type = None
        try:
            trigger_type = {"i": "00", "e": "01", "s": "10"}[trig[0].lower()]
        except AttributeError:
            raise ValueError(
                f"trig not valid, got {trig}, "
                "options are 'immediate', external', 'self'"
            )

        try:
            lbType = {"f": "00", "t": "01", "r": "10"}[lb[0].lower()]
        except AttributeError:
            raise ValueError(
                f"Invalid lbType, got {lb}, "
                "options are 'forced', 'trigrel', or 'roi'"
            )

        try:
            acq_type = {"r": "0", "p": "1"}[acq[0].lower()]
        except AttributeError:
            raise ValueError(
                f"invalid acq_type, got {acq}: " "options are 'raw', or 'pedsub'"
            )

        try:
            pedestals_type = {"z": "00", "c": "01", "r": "11"}[ped[0].lower()]
        except AttributeError:
            raise ValueError(
                f"invalid pedestals_type, got {ped}: "
                "options are 'zero', 'chanoffset', or 'ramaddr'"
            )

        return trigger_type, lbType, acq_type, pedestals_type

    def stop_readout(self):
        """Toggles the "stopacq" signal on the readout module.

        It's equivalent of asking it nicely to stop reading.
        """
        self._write_control_register("stopacq", True)
        self._write_control_register("stopacq", False)

    def toggle_trigger(self, cycles: int = None):
        """Toggles the ext trigger using software.

        Args:
            cycles (int): number of cycles to hold the trigger high for.
                If not provided, a default value is used.
        """
        if cycles is None:
            cycles = self._trigger_wait_cycles
        if not isinstance(cycles, int):
            raise TypeError(f'"cycles" must be an int')
        if cycles < 1:
            raise ValueError(f'"cycles" must be at least 1, got {cycles}')
        if cycles > 2**16 - 1:
            raise ValueError(
                f'"cycles" must be at most {2**16 - 1} (16-bits), got {cycles}'
            )
        self._send_command(f"C000{cycles:04X}")

    def reset_board(self):
        """Try and reset the board.

        In case the FPGA get stuck this can help reset the state.
        """
        time.sleep(0.02)
        self.digital_reset()
        time.sleep(0.02)
        self.sysrst()

    def digital_reset(self):
        """Toggles the "reset" port on the readout module.

        Forcibly returns the chip to default state.
        """
        self._write_control_register("digrst", True)
        self._write_control_register("digrst", False)

    def clear_buffer(self):
        """Clears the UART buffer."""
        if self.board.using_new_backend:
            device = ConnectionManager(self.board).device
            if device.type in NON_UDP:
                device.clear_buffers()
        else:
            self.board.connection.reset_input_buffer()

    def sysrst(self):
        """Toggles the sysrst pin, which resets the digital portion of the chip"""
        self._write_control_register("sysrst", True)
        self._write_control_register("sysrst", False)

    #############################################################################
    # Random Stuff, may or maynot work... Below functions need to be validated.
    #############################################################################

    def read_scalers(self, channels: Iterable[int] = None) -> list[int]:
        """Reads and returns all the digital scalar registers.

        The scalar registers are in two locations, one is the scal{ch} where the lower bits are
        stored, once that register is read, the scalhigh register is populated with the high bits.

        Args:
            channels (list[int]): list of channels to read scalars for.
                If not provided, scalars for all channels are read.

        Returns:
            list[int]: list of register read results for the channels selected.
                The list will not include values for disabled channels.

        Raises:
            ValueError: if a channel number in the list is invalid.
            TypeError: if channels is not a list or channel number is not an integer.
        """
        total_channels = self._get_total_channels()

        if channels is None:
            channels = range(total_channels)
        self._validate_channels_or_raise(channels)
        self.stop_readout()

        result = self._read_scalars(channels)

        return result

    def _read_scalars(self, channels: Iterable[int]) -> list[int]:
        """Reads the scalars for the given channels."""
        result = []
        for chan in channels:
            scalar = self.read_scalar(chan)
            result.append(scalar)
        return result

    def read_scalar(self, channel: int):
        name = self.get_scal_name(channel)
        scal = self._read_digital_register(name)
        try:
            scalhigh = self._read_digital_register("scalhigh")
        except (KeyError, AttributeError):
            scalhigh = 0
        shift_amt = DigitalRegisters(self.board).registers[name]["bitwidth"]
        scal += scalhigh << shift_amt

        return scal

    def get_scal_name(self, i):
        return f"scal{i}"

    def _read_digital_register(self, name: str, chips: "int | list[int]" = None):
        """Quick access to read digital register value"""
        return DigitalRegisters(self.board, chips).read(name)["value"]

    def _get_total_channels(self) -> int:
        """Returns the total number of channels on the board."""
        return self.board.channels

    def read_firmware_version(self):
        """Read the firmware version.

        This is a control register.
        """
        result = -1
        try:
            result = ControlRegisters(self.board).read("version")["value"]
            LOGGER.debug("Firmware version: %s", result)
        except:
            LOGGER.error("Can't read firmware version")
        return result

    def read_identifier(self) -> str:
        """Read the board identifier."""
        result = -1
        try:
            result = ControlRegisters(self.board).read("identifier")
        except:
            LOGGER.error("Can't read board identifier")
        return result

    def enable_testmode(self, enabled: bool):
        """Enables the test-pattern output.

        Set the fpga in a mode to output a known test-pattern.
        """
        ControlRegisters(self.board).write("fake", enabled)

    def get_available_chips(self) -> list[int]:
        """Get a list of available chip numbers."""
        return list(range(self.board.params.get("num_chips", 1)))

    def set_serial_mode(self, enabled: bool):
        """Turn on or off serial mode.

        Warning: there is limited support in NaluDAQ for making sense
        of the data that comes out of a board when in serial mode, use
        at your own risk.

        Args:
            enabled (bool): True to use the serial interface, or False to
                use the parallel interface (this is the default when
                starting up a board).
        """
        if not isinstance(enabled, bool):
            raise TypeError("Argument must be bool")
        LOGGER.debug("Setting serial mode: %s", "ON" if enabled else "OFF")
        registers_to_write = {
            True: {
                "iomode1": 1,
                "iomode0": 0,
            },
            False: {
                "iomode1": 0,
                "iomode0": 1,
            },
        }[enabled]
        for name, value in registers_to_write.items():
            self._write_control_register(name, value)

    def _write_control_register(self, register, value):
        """wrapper for the Control register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        """
        try:
            ControlRegisters(self.board).write(register, value)
        except (ValueError, TypeError) as error_msg:
            LOGGER.error("Couldn't update control register due to: %s", error_msg)

    def _validate_channels_or_raise(self, channels: list[int]):
        """Raise an error if the value is not a valid list of channels"""
        if not isinstance(channels, list):
            raise TypeError(f"Channels must be a list[int], not {type_name(channels)}")
        if any(not isinstance(c, int) for c in channels):
            raise TypeError("Channels must be a list[int]")
        if any(not 0 <= c < self.board.channels for c in channels):
            raise ValueError("One or more channels is out of bounds")

    def _send_command(self, command):
        """Send the given hex command to the board.

        Args:
            command (str): hex command
        """
        if self.board.using_new_backend:
            BoardIoManager(self.board).write(command)
        else:
            self.board.connection.send(command)



"""Module containing base classes for register spaces which are split into multiple chips
-- analog and digital registers.

Commands generated will be wrapped in chip selection commands which will select the appropriate
chips before sending the command, and restore the previous chip selection afterwards. The general
layout of generated pseudo-commands might look like:

- Select chips
- Write address
- Restore chips

This multichip stuff is a bit tricky. The basic idea is that you can write to multiple chips
at once, and the commands generated will be optimized to minimize the number of commands
sent to the board. A true multichip broadcast (same command sent to all boards) is not always
possible because registers living on the same address may have different values for
different chips. In this case, any differing commands will have their own selection logic.
However, a command which is the same for all chips will be broadcast to all chips at once.
A command which ends up being split into multiple commands will look like this:

- Select some chips
- Write address
- Select other chips
- Write address
...
- Restore chips

Some consequences of this implementation are that doing something like `DigitalRegisters.write_addr(addr)`
might generate a very long command, while `DigitalRegisters.write_addr(addr, value)` will generate
a very short command. This is because the former MAY generate a command for each chip.

Multichip reads are currently not supported because we don't have a parser for them, or
consistent firmware support. For now, reads are only supported from one chip at a time.
"""
import functools
import typing
from collections import defaultdict

from . import _common as helpers
from .chip_selection import wrap_command
from .registers import ReadRegistersABC, WriteRegistersABC


def raise_if_multiple_chips(fn):
    """Raise a NotImplementedError if the number of chips selected is greater than 1."""

    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        if len(self.chips) > 1:
            raise NotImplementedError("Cannot read from multiple chips at once")
        return fn(self, *args, **kwargs)

    return wrapper


class ChipRegisters(WriteRegistersABC, ReadRegistersABC):
    def __init__(self, board, register_type: str, chips: "int | list[int]" = None):
        super().__init__(board, register_type)
        self.chips = chips

    @property
    def chips(self) -> list[int]:
        """Get/set the currently selected chips.

        Can be set to a single chip number, a list of chip numbers, or None (all chips).
        """
        return self._chips

    @chips.setter
    def chips(self, chips: "int | list[int]"):
        if isinstance(chips, int):
            chips = [chips]
        elif isinstance(chips, typing.Iterable):
            chips = list(chips)
        elif chips is None:
            chips = list(range(self.board.available_chips))
        else:
            raise TypeError(
                f"chips must be int, list[int], or None, not {type(chips).__name__}"
            )
        if len(chips) == 0:
            raise ValueError("Need to write to at least one chip")
        if any(not 0 <= chip < self.board.available_chips for chip in chips):
            raise ValueError("Invalid chip index")
        self._chips = chips

    # ---------------------------------------------------------------------
    # Write
    # ---------------------------------------------------------------------
    def generate_write(self, name: str, value: int = None) -> str:
        """Generate a write command for the given register.

        This will generate a sequence of commands on multichip boards.
        Since different chips can have different values on neighboring registers,
        multiple commands may be required to write to a single register on different chips
        to avoid inadvertently overwriting neighboring registers on other chips;
        in other words, true broadcast is not possible in the general case.

        Args:
            name (str): the name of the register
            value (int): the value to write to the register. If None,
                the current value of the register will be written.

        Returns:
            str: the command to write the given value to the given register
        """
        self._validate_name_or_raise(name)
        name = name.lower()
        if value is not None:
            bitwidth = self.registers[name]["bitwidth"]
            self._validate_value_or_raise(value, bitwidth)
        return self._wrap_commands(
            [self._generate_write_name_atom(chip, name, value) for chip in self.chips]
        )

    def generate_write_addr(self, addr: "int | str", value: int = None) -> str:
        """Generate a write command for the given address.

        To avoid inadvertently overwriting neighboring registers on other chips,
        multiple commands may be required to write to a single address on different chips;
        in other words, true broadcast is not possible in the general case.

        Args:
            addr (int | str): the address to write to.
            value (int): the value to write to the address. If None,
                the current value of the register will be written.

        Returns:
            str: the command to write the given value to the given address
        """
        if isinstance(addr, int):
            addr = f"{addr:X}"
        if not isinstance(addr, str):
            raise TypeError("Address should be either a hex string or integer")
        self._validate_addr_or_raise(addr)

        return self._wrap_commands(
            [self._generate_write_addr_atom(chip, addr, value) for chip in self.chips]
        )

    def _generate_write_addr_atom(
        self,
        chip: int,
        addr: "int | str",
        value: int = None,
    ) -> str:
        """Generate a write command "atom" for an address for a single chip.

        Args:
            chip (int): Chip number.
            addr (int | str): Address to write to.
            value (int): Value to write to the address. If None, the value from the
                software register will be used.

        Raises:
            ValueError: if the address is invalid

        Returns:
            str: Write command for the given address and chip.
        """
        addr = addr.zfill(self._width_addr).upper()
        if value is None:
            try:
                value = self._addr_value(addr, chip)
            except AttributeError:
                raise ValueError(f"No value found for address {addr}")
        self._validate_value_or_raise(value, bitwidth=self._width_val * 4)
        return self._generate_write_atom(addr, value)

    def _generate_write_name_atom(self, chip: int, name: str, value: int) -> str:
        """Generate a write command "atom" for a register for a single chip.

        Args:
            chip (int): Chip number.
            name (str): Name of the register to write to.
            value (int): Value to write to the register.

        Returns:
            str: Write command for the given register and chip.
        """
        name = name.lower()
        addr = self._get_addr_from_name_or_raise(name)
        if value is not None:
            addr_val = int(self._addr_value(addr, chip))
            value = helpers.substitute_value(addr_val, value, self.registers[name])
        return self._generate_write_addr_atom(chip, addr, value)

    def _wrap_commands(self, commands: list[str]) -> str:
        """Collapse a list of commands into a single command with chip selection.

        Condenses commands which are identical except for the chip number into a single
        broadcasted write.

        Args:
            commands (list[str]): list of commands to collapse.

        Returns:
            str: the collapsed command
        """
        if len(commands) != len(self.chips):
            raise ValueError(
                f"Expected {len(self.chips)} commands, but got {len(commands)} commands"
            )
        collapsed = defaultdict(list)
        for chip, command in zip(self.chips, commands):
            collapsed[command].append(chip)
        command = "".join(
            wrap_command(
                self.board,
                command,
                chips,
                restore=(i == len(collapsed) - 1),
            )
            for i, (command, chips) in enumerate(collapsed.items())
        )
        return command

    def set(self, name: str, value: int):
        values = self.registers[name.lower()]["value"]
        for chip in self.chips:
            values[chip] = value

    def set_addr(self, addr: str, value: int):
        self._validate_value_or_raise(value, bitwidth=self._width_val * 4)
        for name, reg in self.list_registers().items():
            if reg["address"].lower() != addr.lower():
                continue
            val = helpers.full_to_partial_value(value, reg)
            self.set(name, val)

    def _addr_value(self, addr: str, chip: int) -> int:
        """Get the current value of the software register at the given address.

        Args:
            addr (str): the address in hex string format
            chip (int): the chip number

        Returns:
            int: the current value of the register
        """
        regs = self.registers_on_address(addr)
        return functools.reduce(
            lambda value, reg: (
                value | helpers.partial_to_full_value(reg, reg["value"][chip])
            ),
            regs.values(),
            0,
        )

    # ---------------------------------------------------------------------
    # Read
    # ---------------------------------------------------------------------
    def generate_read_addr(self, addr: "int | str") -> str:
        command = self._generate_read_atom(addr)
        command = wrap_command(self.board, command, self.chips)
        return command


_SINGLE_CHIP_FNS = [
    "read",
    "read_all",
    "read_many",
    "generate_read",
    "generate_read_addr",
]
for _name in _SINGLE_CHIP_FNS:
    _fn = getattr(ChipRegisters, _name)
    _fn = raise_if_multiple_chips(_fn)
    setattr(ChipRegisters, _name, _fn)



import http

import requests

from .exceptions import BackendError, HttpError


class HttpClient:
    def __init__(self, addr: tuple[str, int]) -> None:
        """Simple wrapper for sending HTTP requests to the backend.

        Args:
            addr (tuple[str, int]): address of the client.
        """
        host, port = addr
        self._address = addr
        self._prefix = f"http://{host}:{port}"
        self._session = requests.Session()
        self._timeout_sec = 2

    @property
    def address(self) -> tuple[str, int]:
        """Get the server address as (ip, port)"""
        return self._address

    @property
    def host(self) -> str:
        """Get the host IP of the server"""
        return self._address[0]

    @property
    def port(self) -> int:
        """Get the port number of the server"""
        return self._address[1]

    @property
    def url(self) -> str:
        """Get the URL used to reach the server"""
        return f"http://{self.host}:{self.port}/"

    def reachable(self) -> bool:
        """Checks if the server can be reached."""
        try:
            self.get("/connection/info")
            return True
        except BackendError:
            return False

    def put(self, route: str, *, params=None, json=None) -> requests.Response:
        """Send a PUT request to the server."""
        return self._request("put", route, params=params, json=json)

    def put_binary(self, route: str, *, data: bytes, params=None) -> requests.Response:
        """Send a PUT request with binary content to the server."""
        return self._request(
            "put",
            route,
            params=params,
            data=data,
            headers={"Content-Type": "application/octet-stream"},
        )

    def post(self, route: str, *, params=None, json=None) -> requests.Response:
        """Send a POST request to the server."""
        return self._request("post", route, params=params, json=json)

    def post_binary(self, route: str, *, data: bytes, params=None) -> requests.Response:
        """Send a POST request with binary content to the server."""
        return self._request(
            "post",
            route,
            params=params,
            data=data,
            headers={"Content-Type": "application/octet-stream"},
        )

    def delete(self, route: str, *, params=None, json=None) -> requests.Response:
        """Send a DELETE request to the server."""
        return self._request("delete", route, params=params, json=json)

    def get(self, route: str, *, params=None, json=None) -> requests.Response:
        """Send a GET request to the server."""
        return self._request("get", route, params=params, json=json)

    def get_json(self, route: str, *, params=None, json=None) -> dict:
        """Send a GET request to the server and parse the response from JSON."""
        return self.get(route, params=params, json=json).json()

    def get_binary(self, route: str, *, params=None, json=None) -> bytes:
        """Send a GET request to the server and return the raw binary response content"""
        return self.get(route, params=params, json=json).content

    def _request(
        self,
        method: str,
        route: str,
        **kwargs,
    ) -> requests.Response:
        """Sends a generic request to the backend.

        Args:
            method (str): HTTP method as a string (e.g. "post")
            route (str): route for the request
            **kwargs: any arguments accepted by `requests.request`

        Raises:
            HttpError: a generic HTTP error if the status code is not OK.

        Returns:
            requests.Response: the response from the backend.
        """
        extra_args = {"timeout": self._timeout_sec}
        kwargs |= extra_args
        url = self._format_route(route)
        response = self._session.request(method, url=url, **kwargs)
        status = http.HTTPStatus(response.status_code)
        if status == http.HTTPStatus.OK:
            return response
        try:
            error = HttpError(status, response.json())
        except Exception as e:
            raise BackendError("The server sent a malformed response") from e
        error.raise_typed()

    def _format_route(self, route: str) -> str:
        """Get the full URL for a route."""
        if route.startswith("/"):
            route = route[1:]
        return self.url + route



"""i2c controller, talks to the i2c modules o the board.

The i2c communication is both board and chip independent.
"""
from logging import getLogger

from naludaq.communication.i2c_registers import I2CRegisters

LOGGER = getLogger(__name__)
__all__ = [
    "scanI2cDevices",
    "sendI2cCommand",
    "readI2cResponse",
    "readI2cReg",
]


def scanI2cDevices(board) -> list:
    """Checks active devices on the i2c buss.

    Sends out a call to all addresses over the i2c buss and returns
    devices that return an ACK.

    Returns:
        list of devices which returns an ACK.
    """
    response = list()
    for i in range(0, 256, 2):
        addr = hex(i).lstrip("0x").zfill(2)
        words = list()
        if sendI2cCommand(board, addr, words):
            if readI2cResponse(board)[0][-1] == "0":
                response.append(hex(i))

    return response


def sendI2cCommand(board, addr, words, check_ack=False) -> bool:
    """Send commands to the FPGA to rout to i2c devices.

    An abstraction to send an i2c command to GenericWrite.vhd so you don't have to do the dumb
    parsing part input 'words' should be an ordered list of commands in either hex or binary
    address assumes 7 bits in either hex or binary
    and this is always a write.
    be really careful editiing this, because the ordering is both very very important and
    super weird.

    Args:
        addr(str, int): Address of the i2c device, either 7 bits or hex.
        words(list): Words to send. (8-bits)
        board: targeted board.
        check_ack: reads back the response to see if there was an ack

    Returns:
        True if command is sent, False otherwise
    """

    if not isinstance(addr, (str, int)):
        raise TypeError("addr should be a string or int")
    if not isinstance(words, list):
        raise TypeError("words should be a list")

    # LOGGER.debug("sending %s to i2c addr:%s", words, addr)
    if len(words) > 8:
        LOGGER.info("max number of sequential words is 8")
        return False

    try:
        if not isinstance(addr, int):
            addr = _create_i2c_address(addr)
    except Exception as error_msg:
        LOGGER.debug("Couldn't create i2c addr due to: %s", error_msg)
        return False

    # Update the Registers on the board
    try:
        _update_i2c_registers(board, addr, words)
    except Exception as error_msg:
        LOGGER.debug("Couldn't create i2c command due to: %s", error_msg)
        return False

    try:
        I2CRegisters(board).transmit_command()
    except Exception as error_msg:
        LOGGER.debug("Couldn't send i2c command due to: %s", error_msg)
        return False

    if check_ack:
        return all(_check_ack(board, addr, words))

    return True


def _check_ack(board, addr, words):
    """
    Reads back the bitstream from the device and determines if there was an n/ack

    Returns a bool array.  False for NACK, True for ACK

    if addr is *odd*, then it is a write, and only the addr is checked
    """
    response = readI2cResponse(board)
    result = list()
    num_words = len(words) * (addr % 2) + 1
    for iword in range(num_words):
        if response[iword][-1] == "1":
            result.append(False)
            LOGGER.debug("Found a NACK")
        else:
            result.append(True)

    return result


def _create_i2c_address(addr) -> int:
    """Takes a either a 7bit or a hex address and convert into a word address.

    Args:
        addr(str): either a 7bit or a hex address

    """
    if len(addr) == 2:
        addr = int(addr, 16)
    elif len(addr) == 7:
        addr = int(addr, 2)
    else:
        raise ValueError(f"I2c address: {addr} not understood")
    return addr


def _update_i2c_registers(board, addr, words) -> str:
    """Update the I2C registers using the given command parameters.

    Args:
        addr(int): Address of the i2c device in hex.
        words(str): Words to send to the device.
    """

    if len(words) > 8:
        raise ValueError("Too many words, maximum 8 words, provided %s", len(words))

    if addr is None:
        raise TypeError("The provided address type is None.")

    allbits = ""
    for word in words[::-1]:
        if len(word) == 2:
            word = "{0:b}".format(int(word, 16)).zfill(8)
        elif len(word) != 8:
            raise ValueError("i2c word {word} not understood")

        allbits += word

    registers = {
        "i2c_words": len(words),
        "i2c_addr": addr,
    }

    allbits = allbits.ljust(64, "0")
    for reg in range(0, 4):
        name = "i2c_data" + str(3 - reg)
        pair = allbits[reg * 16 : (reg + 1) * 16]
        value = int(pair[8:] + pair[:8], 2)
        registers[name] = value
    I2CRegisters(board).write_many(registers)


def readI2cResponse(board):
    """Read the i2c responses from the registers on the FPGA.

    The i2c module now writes out the SDA value on each high SCL cycle to some registers
    You should be able to parse N/ACKs out of this, as well as responses from devices!
    registers are NGPR+8 through NGPR+11

    Returns:
        List with the responses.
    """
    reg_names = [f"response{i}" for i in range(4)]
    response_regs = I2CRegisters(board).read_many(reg_names)
    phrase = "".join(f"{r:016b}" for r in response_regs)

    phrase = phrase[::-1]
    phrases = []
    if phrase:
        for i in range(0, 7):
            phrases.append(phrase[i * 9 : (i + 1) * 9])

    return phrases


def readI2cReg(board, wraddr, reg):
    """Reads a register from an i2c device

    Basically just sends the address and the reg location, then does the same but with a read

    """

    rdaddr = "{0:02x}".format(int(wraddr, 16) + 1)

    sendI2cCommand(board, wraddr, [reg])
    sendI2cCommand(board, rdaddr, ["FF"])
    response = readI2cResponse(board)
    LOGGER.debug("I2c: %s", response)
    response_int = int(response[1], 2) >> 1

    LOGGER.debug(
        "Register:%s, %s %s %s (ACK,nACK)=%s",
        reg,
        response[1],
        response_int,
        hex(response_int),
        response[1][-1],
    )

    return response_int



"""Center the timescale of an event around the trigger value.

DESCRIPTION:
Finds the first intersections between the graph and the trigger value.
Then creates a new timescale with the

"""
from copy import deepcopy
from logging import getLogger

import numpy as np

LOGGER = getLogger(__name__)


class PostProcessing:
    def __init__(self):
        # self.flags
        self.postprocessors = list()

    def process_event(self, event):
        for processor in self.postprocessors:
            processor(event)


def trim_event(evt, keep_left=False, keep_right=False):
    """Prunes all channels on both left and right side.

    This function is destructive, it will alter the inputed event permanently.


    Args:
        evt(Event): prune target
        keep_left(int): samples to keep before the trigger
        keep_right(int): samples to keep after the trigger

    """
    # should be negative and the amount to cut is based on this.
    first_sample = int(evt["time"][0][0])
    last_sample = int(evt["time"][0][-1])

    remove_left = 0
    if keep_left is not False:
        if abs(first_sample) >= keep_left > -1:  # first sample should be negative
            remove_left = abs(first_sample) - keep_left

    remove_right = -1
    if keep_right is not False:
        if last_sample >= keep_right > -1:
            remove_right = (
                abs(first_sample) + keep_right
            )  # keep to the right of the trigger

    output = []
    for chan in range(len(evt["data"])):
        output.append(evt["data"][chan][remove_left:remove_right])  # Trim all channels
    evt["data"] = None
    evt["data"] = output

    output = None
    try:
        output = evt["time"][:, remove_left:remove_right]
        evt["time"] = None
        evt["time"] = output
    except TypeError as error_msg:
        raise TypeError(
            f"{str(error_msg)} types: "
            f" {str(type(evt['time']))}, {str(type(evt['data']))}"
            f" Values: {str(remove_left)}, {str(remove_right)}"
        )

    return evt


class CenterOnTrigger:
    def __init__(self, raising_edge=True):

        self.raising_edge = raising_edge

    def find_intersection(self, evt, trigger_value, chan=0):
        """Find the first intersection between data and trigger_value.

        Args:
            evt(Event): Event to find the intersection in
            trigger_value(int): value to intersect
            chan(int): Which channel to track?

        Returns:
            the sample number where the data intersects with the trigger_val or 0.
        """
        if getattr(evt, "data", False):
            return 0

        index_intersection = 0
        event_data = np.array(evt["data"][chan], dtype="int")

        try:
            diff = np.diff(np.sign(event_data - trigger_value))
        except (Exception, TypeError) as e_msg:
            LOGGER.debug("something wrong with Center on trigger, %s", e_msg)
            return 0

        coefficient = -1 + 2 * self.raising_edge
        index_intersection = np.argwhere(coefficient * diff > 0).flatten()
        for x in index_intersection:
            return x

        # Can't find the edge? flip the raising edge to falling edge.
        self.flip_edge()

        index_intersection = np.argwhere(coefficient * diff < 0).flatten()

        for x in index_intersection:
            return x

        return 0

    def adjust_time_axis(self, evt, trigger_val, chan=0):
        """Align the timeline based on the trigger.

        find the intersection with the trigger value and center the timescale on the intersection.
        The function updates the passed event.

        Args:
            evt(Event): event to centera
            trigger_val(int): value the board tirggers on.
            chan(int): Channel to find intersction on.
        """
        adjusted_evt = deepcopy(evt)
        flip_pt = self.find_intersection(evt, trigger_val, chan)

        if evt["time"][chan][flip_pt] == 0:
            flip_pt = 0
        try:
            event_time = np.array(evt["time"], dtype=float)
        except ValueError:
            event_time = np.array(evt["time"], dtype=object)

        adjusted_evt["time"] = event_time - flip_pt

        return adjusted_evt

    def flip_edge(self):

        self.raising_edge = not self.raising_edge


class LinearCorrect:
    """Old Bencode to linear correct an event.

    Currently it will only correct on sample at the time.
    Incredibly slow.

    """

    def __init__(self):

        # Linear correction is
        lin_corr_fit = [5.38991698e02, -5.03758060e-01, 6.53423798e-04, -6.60463292e-08]
        self.lin_corr_x = np.arange(0.0, 4096.0)
        self._lin_corr_lin = (
            (lin_corr_fit[3] * self.lin_corr_x**3)
            + (lin_corr_fit[2] * self.lin_corr_x**2)
            + (lin_corr_fit[1] * self.lin_corr_x)
            + lin_corr_fit[0]
        )

    def linear_correction(self, sample):
        """Linear correct a sample."""
        lin_index = np.abs(self._lin_corr_lin - sample).argmin()
        corr = self.lin_corr_x[lin_index]
        return corr


def shift_time_for_forced_mode(event: dict, insert_nan: bool = True):
    """Shifts the time axis to start at the start window. Mainly useful for events
    captured in forced mode.

    If the window labels roll over (end of sampling array is reached in event),
    then the time axis is adjusted to roll over as well. For use in plotting, this
    can cause some ugly jumps, hence the `insert_nan` parameter.

    Warning: as a safety precaution against shifting the axis multiple times,
    this function will ignore events if the time axis does not start at zero.

    Args:
        event (dict): event to work on. The event is modified in-place.
        insert_nan (bool): whether to insert a single `np.nan` into the
            time and data arrays when a discontinuity is created. Useful
            for plotting.

    Returns:
        dict: the same event dict passed in.
    """
    # If the time axis doesn't start at zero, all the processing is invalid
    if int(event["time"][0][0]) != 0:
        return event

    # Disabled channels are empty lists, which cause an error since the data/time arrays are ragged
    _fill_empty_channels_with_nan(event)

    data = np.array(event["data"], dtype=np.float)  # float required for using np.nan
    time = np.array(event["time"], dtype=np.float)
    window_labels = np.array(event["window_labels"])

    # Shift time axis to start at start window
    samples = data.shape[1] // window_labels.shape[1]
    start_sample = event["start_window"] * samples
    time += start_sample

    # If window labels roll over, make sure the time axis rolls over too
    roll_over_loc = np.argwhere(
        window_labels < event["start_window"]
    )  # Doesn't necessarily roll back to zero
    if len(roll_over_loc) != 0:
        rollover_chan, rollover_idx = roll_over_loc[0][0], roll_over_loc[0][1]
        max_winds = window_labels[rollover_chan, rollover_idx - 1] + 1
        time %= max_winds * samples

        if insert_nan:
            time = np.insert(time, rollover_idx * samples, np.nan, axis=1)
            data = np.insert(data, rollover_idx * samples, np.nan, axis=1)

    event["data"] = data
    event["time"] = time

    return event


def _fill_empty_channels_with_nan(event: dict):
    """Fill data/time/window_labels fields with np.nan for channels that are
    disabled to make the arrays non-ragged.
    """
    total_samples = max([len(x)] for x in event["data"])
    total_windows = max([len(x) for x in event["window_labels"]])
    nan_for_samples = np.full(total_samples, np.nan, dtype=np.float)
    nan_for_windows = np.full(total_windows, np.nan, dtype=np.float)

    for chan, chan_data in enumerate(event["data"]):
        if len(chan_data) != 0:
            continue
        event["data"][chan] = nan_for_samples
        event["time"][chan] = nan_for_samples
        event["window_labels"][chan] = nan_for_windows



"""UPACParser

*This parser inverts the 10-bit values as the internal ramp is operating in reverse*
This inversion is around 10-bits which means each value is subtracted from 1023.

Use parser to parse the data from the Hardware.
The boards sends a raw bitstream and it needs to be parsed into events.
The parser class is a tool to parse the raw bitstream of 8-bit chunks
into 16 bit words then extracting the 10-bit data, the header info,
channels and window number data.

This data is then returned in a dictionary.
The raw data is preserved.

"""

import logging

import numpy as np

from naludaq.helpers.exceptions import BadDataError  # IncompatibleParserError,
from naludaq.parsers.parser import Parser

LOGGER = logging.getLogger(__name__)


class UDCParser(Parser):
    """Parser for UDC16 data.

    The dataformat is:
        header: b'10' + "CHIP_ID"(6-bits) (1 byte)
        16 ea: (16 x 8196 bytes = 131_136 bytes)
            - "WINDOW"(8-bit) + "CHANNEL"(8-bit) (2 byte)
            - 4196 samples (each sp 16-bits) (8192 BYTES)
            - CRC(1-byte) + 0xFC (2 bytes)
        footer: 0xCAFE (2 byte)

        131,139 bytes total
    """

    def __init__(self, params: dict):
        """Instantiate the parser using the given parameters.

        Args:
            params (dict): the parameters to use when parsing the data. Make sure
                to only pass in params that match those the data was captured with.
        """
        super().__init__(params)
        self.num_wind_header = params.get("window_headers", 1)
        self.num_wind_footers = params.get("window_footers", 1)
        self._data_mask = params.get("data_mask", 0x03FF)  # 10 bits
        self._chip_mask = params.get("chip_mask", 0x3F)  # 6 bits
        self._channel_mask = params.get("chanmask", 0xFF00)
        self._wind_mask = params.get("windmask", 0x00FF)
        self._channel_shift = params.get("chanshift", 8)
        self.invert_data = True

    @property
    def invert_data(self) -> bool:
        """Enable data inversion.

        If true, inverts each data point to (1023 - data)
        """
        return bool(self._max_data_value)

    @invert_data.setter
    def invert_data(self, val: bool):
        if val:
            # Bitwise XOR of 1's results in subtraction (inversion)
            self._max_data_value = 0x03FF
        else:
            # Bitwise XOR of 0's results in no bit flips
            self._max_data_value = 0x0000

    def parse_digital_data_old(self, in_data) -> dict:
        """Parse UDC data

        Args:
            in_data (dict):

        Returns:
            event dict with parser 'data' field.

        Raises:
            BadDataException if length is more or less than expected or no data in the packet.
        """
        self._validate_input_package(in_data)

        data = in_data.get("rawdata", None)
        if data is None:
            raise BadDataError("No raw_data field in input data")

        channels = self.channels
        samples = self.samples
        windows = self.windows
        chan_mask = self._channel_mask
        chan_shift = self._channel_shift
        wind_mask = self._wind_mask
        data_mask = self._data_mask
        raw_data = np.frombuffer(data[1:], dtype="<H")
        max_data_value = self._max_data_value

        # All lengths below in words
        start = 0
        step = self.num_wind_header + windows * samples + self.num_wind_footers
        end = len(raw_data) - 1

        curr_event = {
            "chip": data[0] & self._chip_mask,
            "window_labels": np.zeros(shape=(channels, windows), dtype=int),
            "data": np.zeros(shape=(channels, windows * samples)),
            "start_window": int(raw_data[0] & wind_mask),  # first word
            "footers": [],
        }

        for chidx in range(start, end, step):
            window = int(raw_data[chidx] & wind_mask)
            chan = int((raw_data[chidx] & chan_mask) >> chan_shift)
            # Remaining window numbers are always in order
            window_labels = [int(x % windows) for x in range(window, window + windows)]

            footer = raw_data[chidx + step]
            # bitwise XOR is used flip the bits without assuming max values and data_mask
            # bitwise XOR is also faster then using subtraction.
            curr_event["data"][chan] = max_data_value ^ (
                raw_data[chidx + 1 : chidx + step - 1] & data_mask
            )
            curr_event["window_labels"][chan] = window_labels
            curr_event["footers"].append(footer)

        return curr_event

    def _validate_input_package(self, in_data):
        """Validate the input data.

        Raises:
            BadDataException if:
            - there is no data.
            - there is not a en event amount of bytes, since 8-bits are combined to 16-bit words.
            - the data is not divisable by sample_num + window_headers.
        """
        if not isinstance(in_data, dict):
            raise TypeError("raw_data is not a dict it's a %s", type(in_data))

        if "rawdata" not in in_data or in_data["rawdata"] == b"":
            raise BadDataError("No Data Found")

        # Edge case, data doesn't contain valid words.
        # The UDC dataformat has a 1 byte header, making the amount uneven
        if len(in_data["rawdata"]) % 2 != 1:
            raise BadDataError(
                f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
                f"length is not an even amount of words, "
                f"len {len(in_data['rawdata'])}"
            )



"""
The si5341 is a good clock that we use a lot

This is all the functions that can be used to control its features

Ben Rotter
"""
import logging
import time

import naludaq.communication.i2c as i2c
from naludaq.controllers.controller import Controller
from naludaq.helpers.exceptions import ClockFileError, RegisterFileError
from naludaq.io.io_manager import load_clockfile

logger = logging.getLogger(__name__)


class Si5341Controller(Controller):
    """Clock controller/programmer

    Program the Si5341 clockchip which is used on all Nalu boards.
    The prorgrammer class doesn't have any public attributes and
    thus doesn't have to be instantiated.

    Example:
    ```
    Si5341Controller(board).program()

    ```

    The board object should contain the command string to program the clock.
    It's also possible to supply the clockfile name when running.

    Example:
    ```
    Si5341Controller(board).program(filename)
    ```

    """

    def __init__(self, board):
        super().__init__(board)

        if board.params.get("si5341_address", None) is None:
            raise RegisterFileError(
                "The loaded Register file doesn't have a si5341_address specified."
            )
        self._address = board.params["si5341_address"]

    def program(self, filename=None, safe_mode=True, verbose=False):
        """Setup the clockspeed of the si5341.

        Can read the commands from a .txt file
        Loads the list of commands given by the ClockBuilder Pro program
        Sent as three words: i2cAddr -> address -> value
        If moving to new page (first byte of addr), need to send an additional "change page" command

        > This function turns off UART logging while running due to the amount of commands sent.

        Args:
            filename: full filepath to the file form Clockbuilder Pro.

        Raises:
            Errors in case the filename is invalid or the clockfile is invalid.
        """
        # if not os.path.isfile(filename):
        #     raise FileNotFoundError(f"File {filename} doesn't exist")

        if filename is None and self.board.clock_data is None:
            raise ClockFileError("No clock file loaded!")

        if filename is None and self.board.clock_data is not None:
            commands = self.board.clock_data

        if filename is not None:
            try:
                commands = load_clockfile(filename)
            except:
                raise

        if isinstance(self._address, int):
            devAddr = hex(self._address)[
                2:
            ]  # what will be seen in the logic analyzer (7 bits + 0)
        else:
            devAddr = self._address

        prevPage = "00"

        logger.info("Using si5341 config file %s", filename)
        logger.debug("Sending command to si5341...")

        uart_logger = logging.getLogger("naludaq.board.connections._UART")
        uart_logger_orig = uart_logger.level
        lvl = logging.DEBUG if verbose else logging.ERROR
        uart_logger.setLevel(lvl)

        for command in commands[1:]:
            if command[0] == "sleep":
                time.sleep(float(command[1]) / 1000.0)
                continue
            fullAddr = command[0][2:]
            page = fullAddr[:2]
            addr = fullAddr[2:]
            hexVal = command[1][2:]
            if safe_mode:
                time.sleep(0.01)

            # logger.debug("fullAddr: %s, page: %s, addr: %s, hexVal: %s", fullAddr, page, addr, hexVal)

            if prevPage != page:
                words = ["01", page]
                if not i2c.sendI2cCommand(self.board, devAddr, words, check_ack=False):
                    logger.info("No ACK recieved from Si5341")
                    return False
                logger.debug("words: %s", words)
                prevPage = page
                if safe_mode:
                    time.sleep(0.01)

            words = [addr, hexVal]

            # logger.debug("words: %s", words)
            if not i2c.sendI2cCommand(self.board, devAddr, words, check_ack=False):
                logger.info("No ACK recieved from Si5341")
                return False

        uart_logger.setLevel(uart_logger_orig)
        logger.info("si5341 init complete.")


def program(board, filename=None):
    """Program the Si5341 clockchip

    The board needs to have registers file loaded
    and preferable a clockfile loaded.

    Args:
        board: good old board
        filename (filepath): Full filepath to the clockfile, optional

    """
    Si5341Controller(board).program(filename)



from os.path import exists

import h5py


class LookupTable:
    """
    Class for accessing/storing lookup table data
    Lookup table can be accessed straight from the class
    File is opened on initialization

    Example:
        with LookupTable('filename', 'r+') as lookup_table:
            lookup_table[0,0,:] = 23
            data = lookup_table[0,0,0]
            lookup_table[(2,0,0)] = 6
            captured_data = np.array(lookup_table.captured_data)
            times_captured = np.array(lookup_table.times_captured)
    """

    def __init__(self, file_name: str, file_mode: str):
        """Sets up class to access lookup table from file

        Args:
            file_name (str): Name of lookup table file
            file_mode (str): 'r' Read only, 'r+' Read/write
        """
        if not exists(file_name):
            raise FileNotFoundError("Lookup table file does not exist")
        self.file_name = file_name
        self.file_mode = file_mode
        self.open()

    @property
    def lookup(self):
        """Dataset containing values used for processing data.
        This is the table modified & filled when interpolating
        data from the lookup table generator.
        """
        self._opened_or_raise()
        if not hasattr(self, "_lookup") or not self._lookup.id:
            self._lookup = self._h5_file["lookup_table"]
        return self._lookup

    @lookup.setter
    def lookup(self, lookup):
        self._lookup = lookup

    @property
    def captured_data(self):
        """Dataset containing raw data captured from a dac sweep.
        Modified when adding data to the file using the lookup table
        generator functions.
        """
        self._opened_or_raise()
        if not hasattr(self, "_captured_data") or not self._captured_data.id:
            self._captured_data = self._h5_file["captured_data"]
        return self._captured_data

    @captured_data.setter
    def captured_data(self, captured_data):
        self._captured_data = captured_data

    @property
    def times_captured(self):
        """Number of times a certain entry in the lookup table is
        captured. Mainly used to determine the average value of
        entries captured more than once.
        """
        self._opened_or_raise()
        if not hasattr(self, "_times_captured") or not self._times_captured.id:
            self._times_captured = self._h5_file["times_captured"]
        return self._times_captured

    @times_captured.setter
    def times_captured(self, times_captured):
        self._times_captured = times_captured

    @property
    def settings(self):
        self._opened_or_raise()
        if not hasattr(self, "_settings") or not self._settings.id:
            self._settings = _settings_to_dict(self._h5_file["settings"])
        return self._settings

    def open(self):
        if not self.is_open:
            self._h5_file = h5py.File(self.file_name, self.file_mode)

    def close(self):
        self._h5_file.close()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def __getitem__(self, key):
        self._opened_or_raise()
        return self.lookup[key]

    def __setitem__(self, key, value):
        self._opened_or_raise()
        self.lookup[key] = value

    @property
    def is_open(self) -> bool:
        """Checks if h5 file is opened"""
        return bool(getattr(self, "_h5_file", False))

    def _opened_or_raise(self):
        """Checks if h5 file is opened, if not raise IOError"""
        if not self.is_open:
            raise IOError("Lookup table file is not open")


def _settings_to_dict(value):
    """Converts h5.Group settings to dict"""
    settings = dict()
    if isinstance(value, h5py.Group):
        for key in value.keys():
            settings[key] = _settings_to_dict(value[key])
    for key in value.attrs.keys():
        settings[key] = value.attrs[key]
    return settings



"""
WIP module for working with AODS or other boards through analog debug mode

Kenneth Lauritzen
"""
import time

from naludaq.communication import ControlRegisters
from naludaq.controllers.controller import Controller


class AnalogDebugController(Controller):
    def analog_debug_enable(self):
        """Sets up the control registers to allow for analog debug."""

        ControlRegisters(self.board).write_register_from_name("iomode0", 0)
        ControlRegisters(self.board).write_register_from_name("nrw", 1)
        ControlRegisters(self.board).write_register_from_name("addr_user", True)
        ControlRegisters(self.board).write_register_from_name("data_user", True)

        ControlRegisters(self.board).write_register_from_name("regclr", 1)
        time.sleep(0.5)
        ControlRegisters(self.board).write_register_from_name("regclr", 0)
        return True

    def analog_debug_disable(self):

        ControlRegisters(self.board).write_register_from_name("iomode0", 1)
        ControlRegisters(self.board).write_register_from_name("nrw", 0)
        ControlRegisters(self.board).write_register_from_name("addr_user", False)
        ControlRegisters(self.board).write_register_from_name("data_user", False)

    def direct_register_write(self, addr, data):

        pclk_registers = range(11)
        ControlRegisters(self.board).write_register_from_name("addr", addr)
        ControlRegisters(self.board).write_register_from_name("data", data)

        if addr in pclk_registers:
            ControlRegisters(self.board).write_register_from_name("sel", 1)
            time.sleep(0.5)
            ControlRegisters(self.board).write_register_from_name("sel", 0)
        else:
            ControlRegisters(self.board).write_register_from_name("pclk", 1)
            time.sleep(0.5)
            ControlRegisters(self.board).write_register_from_name("pclk", 0)



"""Answer parsers

Contains a factory function for retrieving an answer parser
based on the board firmware version.

This module was previously located under the naludaq.daq.workers
subpackage, and was relocated to improve the organization of the
naludaq package.
"""


def get_answer_parser(new_firmware=False):
    response = _parse_answers_old
    if new_firmware is True:
        response = _parse_answers

    return response


def _parse_answers(buffer: bytearray) -> dict:
    """Parses a raw binary answer into python data formats.

    Args:
        buffer (bytes): Response to parse.

    Returns:
        parsed answer as a dict expected format.
    """
    header = int.from_bytes(buffer[0:2], byteorder="big", signed=False)
    response = {
        "header": header,
        "read_reg": header & 255,  # only 8 lsb]
        "value": int.from_bytes(buffer[2:4], byteorder="big", signed=False),
        "cmd_id": int.from_bytes(buffer[4:6], byteorder="big", signed=False),
    }

    # end = buff_data[6:8]
    return response


def _parse_answers_old(buffer: bytearray) -> dict:
    """Parses a raw binary answer into python data formats.

    Args:
        buffer (bytes): Response to parse.

    Returns:
        parsed answer as a dict expected format.
    """
    header = int.from_bytes(buffer[0:2], byteorder="big", signed=False)
    response = {
        "header": header,
        "read_reg": header & 255,  # only 8 lsb]
        "value": int.from_bytes(buffer[2:4], byteorder="big", signed=False),
    }

    return response



import datetime
from copy import deepcopy
from pathlib import Path

from naludaq.backend.exceptions import AcquisitionError
from naludaq.backend.managers.base import Manager
from naludaq.backend.models.acquisition import RemoteAcquisition, TemporaryAcquisition


class AcquisitionManager(Manager):
    def __init__(self, board):
        """Utility providing methods for accessing acquisitions on the remote.

        Args:
            context (Context): context used to communicate with the backend.
        """
        super().__init__(board)

    @property
    def current_acquisition(self) -> "RemoteAcquisition | None":
        """Get/set the current output acquisition.

        Set this to `None` to stop data capture on the server side (saves resources).
        """
        name = self.context.client.get_json("/acq/output").get("name", None)
        if name is not None:
            return RemoteAcquisition(self.context, name)
        return None

    @current_acquisition.setter
    def current_acquisition(self, acq: "RemoteAcquisition | None"):
        if acq is not None:
            acq.set_output()
            return
        self.context.client.put("/acq/output")

    def list(self) -> list[RemoteAcquisition]:
        """List all acquisitions in the backend's working directory."""
        acquisitions = self.context.client.get_json("/acq/list")["acquisitions"]
        return [RemoteAcquisition(self.context, name) for name in acquisitions]

    def get(self, name: str) -> RemoteAcquisition:
        """Get an existing acquisition.

        Args:
            name (str): name of the acquisition

        Returns:
            RemoteAcquisition: handle to the acquisition.

        Raises:
            HttpError: if the request failed, or the acquisition does not exist.
        """
        acq = RemoteAcquisition(self.context, name)
        if not acq.exists:
            raise AcquisitionError(f"No such acquisition: {name}")
        return acq

    def create(self, name: str = None, metadata: dict = None) -> RemoteAcquisition:
        """Creates a new acquisition.

        Args:
            name (str): name of the acquisition.
            metadata (dict): metadata to store in the acquisition.
                If not provided, default metadata is built from the board.

        Returns:
            RemoteAcquisition: handle to the acquisition.

        Raises:
            HttpError: if the acquisition already exists or the request failed.
        """
        if name is None:
            name = datetime.datetime.now().strftime("%Y-%m-%d %H-%M-%S.%f")
        if metadata is None:
            metadata = self._build_metadata()
        acq = RemoteAcquisition(self.context, name)
        acq.create(metadata)
        return acq

    def create_temporary(
        self, name: str = None, metadata: dict = None
    ) -> TemporaryAcquisition:
        """Create a temporary acquisition.

        The temporary acquisition should be used as a context manager
        to ensure the acquisition is deleted from the backend when it
        is no longer needed.

        Args:
            name (str): an optional name for the acquisition.
            metadata (dict): optional metadata to include.

        Returns:
            TemporaryAcquisition: handle to the temporary acquisition.
        """
        if name is None:
            name = datetime.datetime.now().strftime("%Y-%m-%d %H-%M-%S.%f")
        if metadata is None:
            metadata = self._build_metadata()
        return TemporaryAcquisition(self.context, name, metadata)

    def _build_metadata(self) -> dict:
        """Create default metadata from the board"""
        return _normalize_metadata(
            {
                "model": self.board.model,
                "params": deepcopy(self.board.params),
                "registers": deepcopy(self.board.registers),
            }
        )

    def lengths(self) -> dict[RemoteAcquisition, int]:
        """Get the lengths of all acquisitions"""
        return {acq: details["len"] for acq, details in self.show_all(len=True).items()}

    def paths(self) -> dict[RemoteAcquisition, Path]:
        """Get the paths of all acquisitions"""
        return {
            acq: Path(details["path"])
            for acq, details in self.show_all(path=True).items()
        }

    def sizes(self) -> dict[RemoteAcquisition, int]:
        """Get the total sizes in bytes of all acquisitions"""
        return {
            acq: details["total_size"]
            for acq, details in self.show_all(total_size=True).items()
        }

    def show_all(
        self,
        *,
        len: bool = False,
        path: bool = False,
        metadata: bool = False,
        chunk_count: bool = False,
        total_size: bool = False,
    ) -> dict[RemoteAcquisition, dict]:
        """Get select information about all acquisitions.

        Args:
            len (bool): fetch the length of each acquisition.
            path (bool): fetch the path of each acquisition.
            metadata (bool): fetch the metadata of each acquisition.
            chunk_count (bool): fetch the number of chunks in each acquisition.
            total_size (bool): fetch the total size of each acquisition.

        Returns:
            dict[RemoteAcquisition, dict]: a dictionary mapping acquisitions to the request information.
        """
        response = self.context.client.get_json(
            "/acq/show-all",
            params={
                "len": int(len),
                "path": int(path),
                "metadata": int(metadata),
                "chunk_count": int(chunk_count),
                "total_size": int(total_size),
            },
        )
        return {
            RemoteAcquisition(self.context, name): details
            for name, details in response.items()
        }


def _normalize_metadata(d: dict) -> dict:
    """Substitute keys/values that are not JSON-encodable with values that are."""

    def _normalize_key(k: object):
        if isinstance(k, int):
            k = str(k)
        return k

    output = {}
    for k, v in d.items():
        k = _normalize_key(k)
        if isinstance(v, dict):
            v = _normalize_metadata(v)
        elif isinstance(v, datetime.datetime):
            v = v.timestamp()
        elif isinstance(v, bytes):
            v = v.hex()
        elif isinstance(v, Path):
            v = str(v)
        output[k] = v
    return output



"""Utilities for accessing acquisitions.

There are a few acquisition classes, which each provide a different way to access acquisitions:

- RemoteAcquisition - access an acquisition through the server.
- LocalAcquisition - access an acquisition in memory.
- DiskAcquisition - access an acquisition loaded from disk.
- TemporaryAcquisition - same as a remote acquisition, but only exists for as long as you need it.

Each of these types is useful in different situations, and provide roughly the same basic interface.
It is recommended to limit use of the RemoteAcquisition type to only where it is necessary, as it is
by far the slowest and most resource-intensive.
"""
import copy
import gzip
import logging
import mmap
import os
import pickle
import struct
from collections import defaultdict, deque
from copy import deepcopy
from pathlib import Path
from typing import Iterator

import numpy as np
import yaml

from naludaq.backend.context import Context
from naludaq.backend.exceptions import AcquisitionError, BackendError
from naludaq.helpers.exceptions import BadDataError
from naludaq.models import Acquisition
from naludaq.parsers import get_parser
from naludaq.tools.waiter import EventWaiter

logger = logging.getLogger("naludaq.backend.acquisition")

AVAILABLE_MISC_DATA_KEYS = ["pedestals", "caldata", "timingcal", "readout_metadata"]

# Disallowed characters for names on windows/linux
# https://stackoverflow.com/questions/4814040/allowed-characters-in-filename
DISALLOWED_ACQ_NAME_CHARACTERS = [
    "\x00",
    "\\",
    "/",
    ":",
    "*",
    "?",
    '"',
    "<",
    ">",
    "|",
    "\n",
]


class RemoteAcquisition:
    _MISC_DATA_CACHE = defaultdict(dict)
    _METADATA_CACHE = {}

    def __init__(self, context: Context, name: str):
        """A utility for using the backend's REST API to access
        acquisition located on the server.

        All operations on the acquisition are performed through HTTP requests.

        Args:
            context (Context): context used to communicate with the server.
            name (str): name of the acquisition.
        """
        self._name = name
        self._context = context

    @property
    def name(self) -> str:
        """Get/set the name of the acquisition.

        Raises:
            ValueError: if the name (when set) is not valid
        """
        return self._name

    @name.setter
    def name(self, name: str):
        if name == self._name:
            return
        _validate_acq_name_or_raise(name)
        try:
            self.context.client.put(
                "/acq/move", params={"source_name": self._name, "dest_name": name}
            )
        except ValueError as e:
            raise ValueError("Invalid destination acquisition name") from e
        except Exception as e:
            raise AcquisitionError("Could not rename acquisition") from e
        self._name = name

    @property
    def context(self) -> Context:
        """The client used to communiate with the remote"""
        return self._context

    def __eq__(self, other: object) -> bool:
        """Checks if two acquisitions represent the same physical acquisition on the same remote"""
        return (
            isinstance(other, RemoteAcquisition)
            and self._name == other._name
            and self._context == other._context
        )

    def __hash__(self) -> int:
        """Compute the hash for this object"""
        return hash(self._name)

    def __len__(self) -> int:
        """Get the number of events in the acquisition"""
        return self.length

    def __getitem__(self, index: "int | slice | list") -> "bytes | list[bytes]":
        """Get events from the remote"""
        if isinstance(index, int):
            return self.raw_event(index)
        if isinstance(index, slice):
            start, stop, step = index.indices(self.length)
            return [self.raw_event(x) for x in range(start, stop, step)]
        if isinstance(index, list):
            return [self.raw_event(x) for x in index]
        raise TypeError("Index must be int or slice")

    def __iter__(self) -> Iterator[dict]:
        """Iterate over raw events"""
        return (self[i] for i in range(len(self)))

    @property
    def length(self) -> int:
        """Get the number of events in the acquisition"""
        info = self._get_info_or_raise(len=True)
        return info["len"]

    @property
    def path(self) -> str:
        """Get the path to the acquisition on the remote"""
        return self._get_info_or_raise(path=True)["path"]

    @property
    def metadata(self) -> dict:
        """Get the acquisition metadata."""
        cache = RemoteAcquisition._METADATA_CACHE
        key = self._cache_key()
        metadata = cache.get(key, None)
        if metadata is None:
            info = self._get_info_or_raise(metadata=True)
            metadata = yaml.safe_load(info["metadata"])
            cache[key] = metadata
        return metadata

    @property
    def params(self) -> dict:
        """Get params from the acq metadata"""
        return self.metadata["params"]

    @property
    def exists(self) -> bool:
        """Checks if this acquisition exists on the remote"""
        try:
            _ = self.length
        except BackendError:
            return False
        return True

    @property
    def readout_metadata(self) -> "dict | None":
        """Get/set the readout metadata for this acquisition"""
        return self._fetch_misc_data("readout_metadata")

    @readout_metadata.setter
    def readout_metadata(self, readout_metadata: "dict | None"):
        self._store_misc_data("readout_metadata", readout_metadata)

    @property
    def pedestals(self) -> "dict | None":
        """Get/set the pedestals for this acquisition"""
        peds = self._fetch_misc_data("pedestals_calibration")
        if isinstance(peds, dict):
            converted_peds = {
                "data": np.array(peds["data"]),
                "rawdata": np.array(peds["rawdata"]),
                "params": peds.get("params", {}),
            }
            # if the data is already a numpy array, this means it was created in the old misc
            # data format. Send it back to the remote to convert it to the new format
            if isinstance(peds["data"], np.ndarray):
                try:
                    self.pedestals = converted_peds
                except Exception as e:
                    logger.error(
                        "Failed to upgrade pedestals to new format", exc_info=e
                    )
            return converted_peds
        return None

    @pedestals.setter
    def pedestals(self, peds: "dict | None"):
        if peds:
            peds = {
                "data": peds["data"].tolist(),
                "rawdata": peds["rawdata"].tolist(),
                "params": peds.get("params", {}),
            }
        self._store_misc_data("pedestals_calibration", peds)

    @property
    def caldata(self) -> "dict | None":
        """Get/set the adc2mv calibration for this acquisition"""
        return self._fetch_misc_data("adc2mv_calibration")

    @caldata.setter
    def caldata(self, adc2mv: "dict | None"):
        self._store_misc_data("adc2mv_calibration", adc2mv)

    @property
    def timingcal(self) -> "list | None":
        """Get/set the timing calibration for this acquisition"""
        return self._fetch_misc_data("timing_calibration")

    @timingcal.setter
    def timingcal(self, timingcal: "list | None"):
        self._store_misc_data("timing_calibration", timingcal)

    def create(self, metadata: dict):
        """Creates the acquisition on the remote.

        Args:
            metadata (dict): metadata to store in the acquisition

        Raises:
            HttpError: if the acquisition exists, or could not be created.
        """
        self._context.client.post(
            "/acq",
            params={"name": self._name},
            json={"metadata": yaml.dump(metadata)},
        )

    def delete(self):
        """Deletes the acquisition on the remote"""
        self.context.client.delete(
            "/acq",
            params={"name": self._name},
        )

    def stream(
        self, timeout: float = 0.5, interval: float = 0.010, start: int = 0
    ) -> Iterator[bytes]:
        """An iterator over events which can be used to stream data during a readout.

        The stream ends when the retrieving the next event times out.
        Note that the iterator only sleeps while at the end of the acquisition;
        for events before the end the stream will run as fast as possible.

        Args:
            timeout (float): timeout in seconds for receiving an event.
            interval (float): the interval in seconds at which to check for events.
            start (int): the event at which to start the stream.
                -1 means to start past the end of the acquisition and wait for a new
                event. This is useful for streaming during a readout.
        """
        index = _resolve_index_or_default(start, len(self) + 1, 0)
        while True:
            waiter = EventWaiter(
                buffer=self,
                amount=index + 1,
                interval=interval,
                timeout=timeout,
            )
            try:
                waiter.start(blocking=True)
            except TimeoutError:
                raise
            try:
                event = self[index]
            except IndexError as e:
                raise TimeoutError("Timed out while waiting for an event") from e
            yield event
            index += 1

    def stream_parsed(
        self,
        timeout: float = 0.5,
        interval: float = 0.010,
        start: int = 0,
        skip_bad=True,
        parser=None,
    ) -> Iterator[dict]:
        """An iterator over parsed events. See the `stream()` documentation for more details.

        Events which cannot be parsed are ignored.

        Args:
            parser (Parser): the parser to use
            timeout (float): timeout in seconds for receiving an event.
            interval (float): the interval in seconds at which to check for events.
            skip_bad (bool): whether to skip invalid events.
        """
        parser = parser or get_parser(self.params)
        for event in self.stream(timeout, interval, start):
            try:
                event = parser.parse(event)
            except BadDataError:
                if not skip_bad:
                    raise
            if "data" in event:
                yield event
            elif not skip_bad:
                raise BadDataError("Got an invalid event")

    def transfer(self, indices: list[int] = None) -> "LocalAcquisition":
        """Transfer a portion of the events from this remote acquisition
        into a `LocalAcquisition`.

        Warning: calling this method is expensive, use wisely.

        Args:
            indices (list[int]): indices of events to transfer. If not
                provided, all events are transferred. This can be
                very costly!

        Returns:
            LocalAcquisition: The local acquisition.
        """
        if indices is None:
            indices = range(len(self))
        events = [self.raw_event(i) for i in indices]
        misc_data = {}
        for name in AVAILABLE_MISC_DATA_KEYS:
            try:
                misc_data[name] = getattr(self, name)
            except:
                pass
        return LocalAcquisition(
            name=self.name,
            metadata=deepcopy(self.metadata),
            events=events,
            misc_data=misc_data,
        )

    def raw_event(self, index: int) -> bytes:
        """Get a raw event at the given index"""
        index = _resolve_index_or_raise(index, len(self))
        rawdata = self.context.client.get_binary(
            "/acq/event",
            params={"acquisition": self._name, "index": index},
        )
        return _build_raw_event_dict(rawdata, index)

    def parsed_event(self, parser, index: int) -> dict:
        """Convenience function for getting a parsed event

        Args:
            parser (Parser): the parser to use
            index (int): index of the event

        Returns:
            dict: the parsed event
        """
        return parser.parse(self.raw_event(index))

    def set_output(self):
        """Use this acquisition as the output for readouts"""
        self.context.client.put(
            "/acq/output",
            params={"name": self._name},
        )

    def _get_info_or_raise(
        self,
        len: bool = False,
        path: bool = False,
        metadata: bool = False,
        chunk_count: bool = False,
        total_size: bool = False,
    ) -> dict:
        """Retrieve information about this acquisition"""
        return self.context.client.get_json(
            "acq/show",
            params={
                "name": self._name,
                "len": int(len),
                "path": int(path),
                "metadata": int(metadata),
                "chunk_count": int(chunk_count),
                "total_size": int(total_size),
            },
        )

    def invalidate_cache(self):
        """Invalidate the misc data cache for this acquisition."""
        cache = RemoteAcquisition._MISC_DATA_CACHE
        try:
            del cache[self._cache_key()]
        except:
            pass  # not an error

    def _fetch_misc_data(self, type: str):
        """Fetch misc data cache, or from the acquisition on miss.

        Args:
            type (str): the type of misc data

        Raises:
            AcquisitionError: if the misc data is invalid

        Returns:
            object: The deserialized misc data
        """
        missing = object()
        key = self._cache_key()
        cache = RemoteAcquisition._MISC_DATA_CACHE
        result = cache[key].get(type, missing)
        if result is missing:
            try:
                result = self._fetch_misc_data_remote(type)
            except AcquisitionError:
                raise
            cache[key][type] = result
        return result

    def _store_misc_data(self, type: str, obj):
        """Serializes and stores the misc data in the cache and acquisition.

        Args:
            type (str): the misc data type name
            obj (object): the misc data

        Raises:
            ValueError: if the misc type is invalid or the object could not
                be serialized.
        """
        cache = RemoteAcquisition._MISC_DATA_CACHE
        cache[self._cache_key()][type] = obj
        try:
            self._store_misc_data_remote(type, obj)
        except ValueError:
            raise

    def _fetch_misc_data_remote(self, type: str):
        """Read misc data from the acquisition.

        Args:
            type (str): the type of misc data

        Raises:
            AcquisitionError: if the misc data is invalid

        Returns:
            object: The deserialized misc data
        """
        try:
            response = self._context.client.get(
                "/acq/misc-data",
                params={"acquisition": self._name, "type": type},
            )
        except ValueError:  # means there is no misc data
            return None
        try:
            return _deserialize_misc_data(response.content)
        except ValueError as e:
            raise AcquisitionError(f"Misc data for {type} is malformed") from e

    def _store_misc_data_remote(self, type: str, obj):
        """Serializes and stores the misc data in the acquisition.

        Args:
            type (str): the misc data type name
            obj (object): the misc data

        Raises:
            ValueError: if the misc type is invalid or the object could not
                be serialized.
        """
        try:
            data = _serialize_misc_data(obj)
        except ValueError:
            raise
        try:
            self._context.client.put_binary(
                "/acq/misc-data",
                data=data,
                params={"acquisition": self._name, "type": type},
            )
        except ValueError as e:
            raise ValueError(f"Invalid misc data name '{type}'") from e

    def _cache_key(self):
        """Get a unique-ish key representing this acquisition to use for the misc data cache"""
        return (self.context.address, self.name)


class TemporaryAcquisition(RemoteAcquisition):
    def __init__(self, context: Context, name: str, metadata: dict):
        """A temporary version of the remote acquisition.

        An acquisition is created on the remote when a `TemporaryAcquisition`
        is instantiated, and deleted when the object is deleted or, if used
        as a context manager, when the context exits.

        Args:
            client (HttpClient): the client
            name (str): name of the acquisition. This must be unique on the remote!
            metadata (dict): metadata to store in the acquisition.
        """
        super().__init__(context, name)
        self.create(metadata)

    def __enter__(self) -> "TemporaryAcquisition":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.exists:
            self.delete()


class LocalAcquisition:
    def __init__(
        self,
        name: str,
        metadata: dict,
        events: list[dict],
        misc_data: dict[str, object],
    ):
        """Local (RAM) copy of a slice of an acquisition.

        Converting a remote acquisition to a local one may be useful in the following cases:
        - the data needs to be retained after the server is shut down/disconnected.
        - the data needs to be modified.
        - the up-front cost of transferring the data is preferred to the cost of
          transferring data while processing it.
        - the acquisition needs to be converted to a legacy format.

        Args:
            name (str): acquisition name
            metadata (dict): acquisition metadata
            events (list[dict]): raw events from the remote
            misc_data (dict[str, object]): misc_data data
        """
        self._name = name
        self._metadata = metadata
        self._events = events
        self._misc_data = misc_data

    @property
    def name(self) -> str:
        """Acquisition name"""
        return self._name

    @property
    def metadata(self) -> dict:
        """Acquisition metadata"""
        return self._metadata

    @property
    def events(self) -> list[dict]:
        """List of events transferred from the remote"""
        return self._events

    @property
    def readout_metadata(self) -> "dict | None":
        """Get/set the readout metadata for this acquisition"""
        return self._misc_data.get("readout_metadata", None)

    @readout_metadata.setter
    def readout_metadata(self, metadata: "dict | None"):
        self._misc_data["readout_metadata"] = metadata

    @property
    def pedestals(self) -> "dict | None":
        """Get/set the pedestals for this acquisition"""
        peds = self._misc_data.get("pedestals", None)
        if isinstance(peds, dict):
            peds = {
                "data": np.array(peds["data"]),
                "rawdata": np.array(peds["rawdata"]),
                "params": peds.get("params", {}),
            }
        return peds

    @pedestals.setter
    def pedestals(self, peds: "dict | None"):
        self._misc_data["pedestals"] = peds

    @property
    def caldata(self) -> "dict | None":
        """Get/set the adc2mv calibration for this acquisition"""
        return self._misc_data.get("caldata", None)

    @caldata.setter
    def caldata(self, adc2mv: "dict | None"):
        self._misc_data["caldata"] = adc2mv

    @property
    def timingcal(self) -> "list | None":
        """Get/set the timing calibration for this acquisition"""
        return self._misc_data.get("timingcal", None)

    @timingcal.setter
    def timingcal(self, timingcal: "list | None"):
        self._misc_data["timingcal"] = timingcal

    def __len__(self) -> int:
        """Number of events in the local copy"""
        return len(self._events)

    def __getitem__(self, index: "int | slice") -> list[dict]:
        """Get a slice of the acquisition events."""
        if isinstance(index, (int, slice)):
            return self._events[index]
        raise TypeError("Index must be int or slice")

    def __iter__(self) -> Iterator[dict]:
        """Iterate over raw events"""
        return (self[i] for i in range(len(self)))

    def raw_event(self, index: int) -> dict:
        """Get a raw event at the given index"""
        index = _resolve_index_or_raise(index, len(self))
        return self.events[index]

    def parsed_event(self, parser, index: int) -> dict:
        """Convenience function for getting a parsed event

        Args:
            parser (Parser): the parser to use
            index (int): index of the event

        Returns:
            dict: the parsed event
        """
        return parser.parse(self.raw_event(index))

    def parse_all(self, parser) -> list[dict]:
        """Parse all events in the local acquisition

        Args:
            parser (Parser): the parser to use

        Returns:
            list[dict]: list of parsed events
        """
        return [parser.parse(event) for event in self]

    def to_legacy_acquisition(self, deepcopy: bool = False) -> Acquisition:
        """Convert to a legacy acquisition object.

        Not all information will be transferred, only the main stuff.
        """
        maybe_copy = lambda x: copy.deepcopy(x) if deepcopy else x

        acq = Acquisition(name=self.name)
        acq.events = deque(maybe_copy(self.events))
        acq.pedestals = maybe_copy(self.pedestals)
        acq.caldata = maybe_copy(self.caldata)
        acq.timingcal = maybe_copy(self.timingcal)
        acq.params = maybe_copy(self.metadata.get("params", {}))
        acq.registers = maybe_copy(self.metadata.get("registers", {}))
        acq.model = acq.params.get("model", None)

        return acq


class DiskChunk:
    INDEX_ENTRY_SIZE = 8

    def __init__(self, index_path: Path, bin_path: Path):
        """Utility for reading chunk files directly from disk.

        The chunk will be opened on instantiation. It is recommended to use
        this object as a context manager to ensure the chunk is closed when
        it is no longer needed.

        Args:
            index_path (Path): path to index file
            bin_path (Path): path to binary file
        """
        if not isinstance(index_path, (Path, str)) or not isinstance(
            bin_path, (Path, str)
        ):
            raise TypeError("Path must be a string or Path object")
        self._index_path = Path(index_path)
        self._bin_path = Path(bin_path)
        self._open = False
        self.open()

    @property
    def metadata(self) -> dict:
        """Get the chunk metadata

        Raises:
            IOError: if the chunk is not open
        """
        self._raise_if_closed()
        return self._metadata

    @property
    def version(self) -> int:
        """Get the chunk format revision number.

        Raises:
            IOError: if the chunk is not open
        """
        self._raise_if_closed()
        return self._version

    def __len__(self) -> int:
        """Get the number of events in the chunk

        Raises:
            IOError: if the chunk is not open
        """
        self._raise_if_closed()
        return len(self._index_mmap) // DiskChunk.INDEX_ENTRY_SIZE

    def __del__(self):
        """Close the chunk file"""
        self.close()

    def __enter__(self):
        """Context manager enter"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Context manager exit"""
        self.close()

    def open(self):
        self._index_file = open(self._index_path, "rb+")
        self._bin_file = open(self._bin_path, "rb+")
        self._index_mmap = mmap.mmap(
            self._index_file.fileno(), os.path.getsize(self._index_path)
        )
        self._bin_mmap = mmap.mmap(
            self._bin_file.fileno(), os.path.getsize(self._bin_path)
        )
        self._open = True

        self._version = self._read_bin_header()["version"]
        self._metadata = self._read_metadata()
        self._parser = get_parser(self._metadata["params"])

    def close(self):
        """Close the chunk file"""
        if self._open:
            self._index_mmap.close()
            self._index_file.close()
            self._bin_mmap.close()
            self._bin_file.close()

    def raw_event(self, index: int) -> dict:
        """Get a raw event from the chunk.

        Args:
            index (int): index of the event in the chunk.

        Returns:
            dict: the raw event

        Raises:
            IndexError: if the index is out of bounds
            IOError: if the chunk is closed
        """
        self._raise_if_closed()
        index = _resolve_index_or_raise(index, len(self))
        start = index * DiskChunk.INDEX_ENTRY_SIZE
        end = start + DiskChunk.INDEX_ENTRY_SIZE
        entry = self._index_mmap[start:end]
        offset, length = struct.unpack("II", entry)
        event = self._bin_mmap[offset : offset + length]
        return _build_raw_event_dict(event, index)

    def parsed_event(self, index: int) -> dict:
        """Get a parsed event from the chunk.

        Uses metadata stored in the chunk file to parse the data.

        Args:
            index (int): index of the event in the chunk.

        Returns:
            bytes: the parsed event

        Raises:
            IndexError: if the index is out of bounds
            IOError: if the chunk is closed
        """
        self._raise_if_closed()
        return self._parser.parse(self.raw_event(index))

    def _read_metadata(self) -> dict:
        """Read metadata from the bin file.

        Raises:
            IOError: if the chunk is closed
        """
        self._raise_if_closed()
        metadata_length = self._read_bin_header()["metadata_length"]
        metadata = self._bin_mmap[8 : 8 + metadata_length]
        return yaml.safe_load(metadata)

    def _read_bin_header(self) -> dict:
        """Read the header from the binary file.

        Returns:
            dict: contains keys "version" and "metadata_length"

        Raises:
            IOError: if the chunk is closed
        """
        self._raise_if_closed()
        version, _, metadata_length = struct.unpack("HHI", self._bin_mmap[:8])  # hello!
        return {
            "version": version,
            "metadata_length": metadata_length,
        }

    def _raise_if_closed(self):
        if not self._open:
            raise IOError("Chunk is closed")


class DiskAcquisition:
    def __init__(self, path: Path, parse: bool = True):
        """Utility for accessing acquisition directly from disk.

        The acquisition will be opened on instantiation. It is recommended to use
        this object as a context manager to ensure the acquisition is closed when
        it is no longer needed.

        Important: the acquisition should not be mutated while this object is alive.
        If the acquisition changes, the DiskAcquisition must be recreated.

        Args:
            path (Path): path to the acquisition folder.

        Raises:
            InvalidAcquisitionError: if the path is not a valid acquisition.
        """
        if not isinstance(path, (Path, str)):
            raise TypeError("Path must be a string or Path object")
        path = Path(path).resolve()
        if not DiskAcquisition._is_acquisition(path):
            raise AcquisitionError("Not a valid acquisition")
        self.parse = parse
        self._root = path
        self._open = False
        self.open()

    @property
    def parse(self) -> bool:
        """Whether to parse events when getting them"""
        return self._parse

    @parse.setter
    def parse(self, value: bool):
        self._parse = value

    def __len__(self) -> int:
        """Get the number of events in the acquisition"""
        return self._event_count

    def __del__(self):
        """Close the acquisition"""
        self.close()

    def __enter__(self) -> "DiskAcquisition":
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __getitem__(self, index: "int | slice") -> "dict | list[dict]":
        """Get one or more parsed event from the acquisition.

        Args:
            index (int | slice): index(es) of the event(s) in the acquisition.

        Returns:
            "dict | list[dict]": the event(s)

        Raises:
            IndexError: if the index is out of bounds
            IOError: if the acquisition is closed
        """
        self._raise_if_closed()
        result = None
        if self._parse:
            fn = self.parsed_event
        else:
            fn = self.raw_event
        if isinstance(index, int):
            index = _resolve_index_or_raise(index, len(self))
            result = fn(index)
        elif isinstance(index, slice):
            result = [fn(i) for i in range(*index.indices(len(self)))]
        elif isinstance(index, list):
            result = [fn(i) for i in index]
        else:
            raise TypeError("Index must be an int or slice")
        return result

    def open(self):
        """Open the acquisition for reading.

        If the acquisition is already open, this is a no-op.
        """
        if self._open:
            return
        self._chunks = self._find_chunks()
        self._event_count = sum(len(chunk) for chunk in self._chunks)
        with open(self._root / "metadata.yml", "r") as f:
            self._metadata = yaml.safe_load(f)
        self._open = True

    def close(self):
        """Close the acquisition"""
        for chunk in self._chunks:
            chunk.close()
        self._chunks = []
        self._open = False

    @property
    def chunks(self) -> list[DiskChunk]:
        """Get a list of chunks in the acquisition"""
        return self._chunks.copy()

    @property
    def metadata(self) -> dict:
        """Get the acquisition metadata"""
        self._raise_if_closed()
        return self._metadata

    @property
    def params(self) -> dict:
        """Get parameters describing the board"""
        return self._metadata.get("params", {})

    @property
    def readout_metadata(self) -> "dict | None":
        """Get/set the readout metadata for this acquisition"""
        return self._fetch_misc_data("readout_metadata")

    @readout_metadata.setter
    def readout_metadata(self, metadata: "dict | None"):
        self._store_misc_data("readout_metadata", metadata)

    @property
    def pedestals(self) -> "dict | None":
        """Get/set the pedestals for this acquisition"""
        peds = self._fetch_misc_data("pedestals_calibration")
        if isinstance(peds, dict):
            peds = {
                "data": peds["data"],
                "rawdata": peds["rawdata"],
                "params": peds.get("params", {}),
            }
        return peds

    @pedestals.setter
    def pedestals(self, peds: "dict | None"):
        if isinstance(peds, dict):
            peds = {
                "data": peds["data"].tolist(),
                "rawdata": peds["rawdata"].tolist(),
                "params": peds.get("params", {}),
            }
        self._store_misc_data("pedestals_calibration", peds)

    @property
    def caldata(self) -> "dict | None":
        """Get/set the adc2mv calibration for this acquisition"""
        return self._fetch_misc_data("adc2mv_calibration")

    @caldata.setter
    def caldata(self, adc2mv: "dict | None"):
        self._store_misc_data("adc2mv_calibration", adc2mv)

    @property
    def timingcal(self) -> "list | None":
        """Get/set the timing calibration for this acquisition"""
        return self._fetch_misc_data("timing_calibration")

    @timingcal.setter
    def timingcal(self, timingcal: "list | None"):
        self._store_misc_data("timing_calibration", timingcal)

    def raw_event(self, index: int) -> bytes:
        """Get a raw event from the acquisition.

        Args:
            index (int): index of the event

        Returns:
            bytes: the raw event data

        Raises:
            IndexError: if the index is out of bounds
            IOError: if the acquisition is not open
        """
        self._raise_if_closed()
        chunk, index_in_chunk = self._map_to_chunk(index)
        return chunk.raw_event(index_in_chunk)

    def parsed_event(self, index: int) -> bytes:
        """Get a parsed event from the acquisition.

        Args:
            index (int): index of the event

        Returns:
            bytes: the raw event data

        Raises:
            IndexError: if the index is out of bounds
            IOError: if the acquisition is not open
        """
        self._raise_if_closed()
        chunk, index_in_chunk = self._map_to_chunk(index)
        return chunk.parsed_event(index_in_chunk)

    def _map_to_chunk(self, index: int) -> tuple[DiskChunk, int]:
        """Map an absolute index to a chunk and a relative index.

        Args:
            index (int): the absolute event index

        Raises:
            IndexError: if the index is out of bounds

        Returns:
            tuple[DiskChunk, int]: the chunk the index corresponds to and the
                event number relative to the chunk.
        """
        for chunk in self._chunks:
            if index < len(chunk):
                return (chunk, index)
            index -= len(chunk)
        raise IndexError("Index out of bounds")

    def _find_chunks(self) -> list[DiskChunk]:
        """Find all chunks located in the acquisition folder.

        Returns:
            list[DiskChunk]: list of chunks
        """
        count = 0
        output = []
        while True:
            index_path = self._root / f"{count}.idx"
            bin_path = self._root / f"{count}.bin"
            if not index_path.exists():
                break
            chunk = DiskChunk(index_path, bin_path)
            output.append(chunk)
            count += 1
        return output

    def _fetch_misc_data(self, type: str):
        """Read misc data from the acquisition.

        Args:
            type (str): the type of misc data

        Returns:
            object: The deserialized misc data

        Raises:
            AcquisitionError: if the misc data is invalid
            IOError: if the acquisition is not open
        """
        self._raise_if_closed()
        try:
            with open(self._root / type, "rb") as f:
                data = f.read()
        except:  # means there is no misc data
            return None
        try:
            return _deserialize_misc_data(data)
        except ValueError as e:
            raise AcquisitionError(f"Misc data for {type} is malformed") from e

    def _store_misc_data(self, type: str, obj):
        """Serializes and stores the misc data in the acquisition.

        Args:
            type (str): the misc data type name
            obj (object): the misc data

        Raises:
            ValueError: if the misc type is invalid or the object could not
                be serialized.
            IOError: if the acquisition is not open
        """
        self._raise_if_closed()
        try:
            data = _serialize_misc_data(obj)
        except ValueError:
            raise
        with open(self._root / type, "wb") as f:
            f.write(data)

    def _raise_if_closed(self):
        if not self._open:
            raise IOError("Acquisition is not open")

    @staticmethod
    def _is_acquisition(root: Path) -> bool:
        """Check if the given path is a valid acquisition"""
        return root.is_dir() and (root / "metadata.yml").exists()


def _build_raw_event_dict(rawdata: bytes, index: int) -> dict:
    """Build the raw event dict from raw raw data.

    Args:
        rawdata (bytes): the raw raw data.
        index (int): the event index

    Returns:
        dict: the event dict
    """
    return {
        "rawdata": rawdata,
        "event_num": index,
        "pkg_num": index,
    }


def _resolve_index_or_raise(index: int, length: int) -> int:
    """Convert a negative index to a positive one.

    Args:
        index (int): the index in the collection
        length (int): the length of the collection

    Raises:
        IndexError: if the index is invalid

    Returns:
        int: the resolved index
    """
    if index < 0 and length > 0:
        index = (length + index) % length
    if not 0 <= index < length:
        raise IndexError(
            f"Index {index} out of bounds for acquisition of length {length}"
        )
    return index


def _resolve_index_or_default(index: int, length: int, default: int) -> int:
    """Convert a negative index to a positive one, or return a
    provided default if the index is invalid.

    Args:
        index (int): the index in the collection
        length (int): the length of the collection
        default (int): the default value

    Returns:
        int: the resolved index, or the default if the index is invalid
    """
    try:
        index = _resolve_index_or_raise(index, length)
    except IndexError:
        index = default
    return index


def _serialize_misc_data(obj) -> bytes:
    """Serialize misc data.

    Args:
        obj (object): the misc data.

    Raises:
        ValueError: if the data could not be serialized.

    Returns:
        bytes: the serialized misc data.
    """
    # Compression level seriously affects how long this takes,
    # don't put it above 3
    try:
        return gzip.compress(pickle.dumps(obj), compresslevel=3)
    except (pickle.PicklingError, gzip.BadGzipFile) as e:
        raise ValueError("Cannot serialize misc data") from e


def _deserialize_misc_data(data: bytes) -> object:
    """Deserialize misc data.

    Args:
        data (bytes): the serialized data.

    Raises:
        ValueError: if the data could not be deserialized

    Returns:
        object: the misc data
    """
    try:
        return pickle.loads(gzip.decompress(data))
    except (pickle.PicklingError, gzip.BadGzipFile, EOFError) as e:
        raise ValueError("Cannot deserialize misc data") from e


def _validate_acq_name_or_raise(name: str):
    """Check whether the given name can be used for an acquisition.

    This isn't a full check since that would be insane. For example
    there are hundreds of unicode control codes that aren't checked.
    This method will only catch the most common naming mistakes a user
    might make.

    Args:
        name (str): acquisition name

    Raises:
        TypeError: if the name is not a string
        ValueError: if the name is zero-length or contains invalid characters
    """
    if not isinstance(name, str):
        raise TypeError("Name must be a string")
    if len(name) == 0:
        raise ValueError("Name cannot have zero length")
    if any(x in name for x in DISALLOWED_ACQ_NAME_CHARACTERS):
        raise ValueError("Name contains an invalid character")
    # Catch ".", "..", and trailing period on Windows
    if name.startswith(".") or name.endswith("."):
        raise ValueError("Name cannot start or end with '.'")
    # Windows thing
    if name.startswith(" ") or name.endswith(" "):
        raise ValueError("Name cannot start or end with space character")



"""Module to convert between acquisition metadata structures of different versions.

This module can be used to upgrade and downgrade metadata formats between
many different versions.

Uses a stack-based approach to "push" and "pop" changes in acq dicts that
appear across versions. Locations of entries in dictionaries are represented
using path strings similar to file system paths.

Supported version strings:
- '0.1': the original Acquisition format. Since this is not a pure dictionary,
    the users of this module will need to handle moving everything into or
    out of a dictionary
- '0.2': placed all metadata into a single dict
- '0.3': remove duplicate model entry

Operations:
- ('move', source_path, dest_path): moves an entry from one location to another
- ('copy', source_path, dest_path): copies an entry from one location to another
- ('set', path, value): stores 'value' in a path
- ('del', path): removes the entry at a path
- ('modify', path, callable): applies a function to an entry, then replaces
    the old entry with the return value of the function.
"""
import ast
import copy
import logging
from collections import deque

from naludaq.helpers.helper_functions import get_package_versions

_LOGGER = logging.getLogger(__name__)


def upgrade_old_acquisition(acq):
    """Upgrades an old acquisition to the latest version

    Args:
        acq (v0.1 Acquisition, dict): acquisition data dict, or
            v0.1 acquisition loaded from a pickle.

    Returns:
        The upgraded acquisition.
    """
    from naludaq.models import Acquisition

    old_version = EARLIEST_VERSION
    if isinstance(acq, dict):
        old_version = acq["metadata"]["file_version"]

    _LOGGER.debug(f"Converting acquisition from {old_version} to {LATEST_VERSION}")

    if type(acq) == deque:
        new_acq = Acquisition()
        new_acq.events.extend(acq)
        return new_acq

    elif old_version == EARLIEST_VERSION:
        naludaq_version, naluscope_version = get_package_versions()

        # Version 0.1 did not use a single dict, need to grab info from a few places
        # and create a fake acq dict for the converter to use
        data = {
            "file_version": EARLIEST_VERSION,
            "naludaq_version": naludaq_version,
            "naluscope_version": naluscope_version,
            "model": acq.info.get("board_info", {}).get("model", None),
            "acq_num": acq._acq_num,
            "created_at": getattr(acq, "created_at"),
            "name": acq._name,
            "notes": getattr(acq, "notes"),
            "numevt": len(acq),
            "info": acq.info,
            "trigger_settings": getattr(acq, "trigger_settings", None),
        }
        data = convert(data, old_version, LATEST_VERSION)
        new_acq = Acquisition()
        new_acq.as_dict().update({"metadata": data})
        new_acq.events.extend(acq)
        new_acq.pedestals = getattr(acq, "_pedestals", None)
        new_acq.caldata = getattr(acq, "_caldata", None)
        new_acq.timingcal = getattr(acq, "_timingcal", None)
        return new_acq

    # > v0.1
    data = convert(acq, old_version, LATEST_VERSION)
    return Acquisition(data=data)


def convert(acq: dict, from_version: str, to_version: str, in_place=False):
    """Converts an acquisition structure from one version to another.

    Args:
        acq (dict): the acquisition dict to work on
        from_version (str): the starting version number
        to_version (str): the ending version number
        in_place (bool): whether to apply the changes directly to the
            dict, or a copy of it.

    Returns:
        A dict in the 'to_version' format.
    """
    _LOGGER.debug(f"Converting metadata from {from_version} to {to_version}")
    try:
        changes = _get_changes_list(from_version, to_version)
    except ValueError as e:
        _LOGGER.error(f"Could not convert metadata: {e}")
        raise

    if not in_place:
        acq = copy.deepcopy(acq)
    _apply_changes(acq, changes)

    return acq


def _get_entry(d: dict, path: str):
    """Finds an entry in a dictionary. Creates dictionaries
    if they do not appear along the path.

    Entries are located using path representation. For example, to reference
    the 'inner' key in  `{'outer': {'inner': 5}}`, the path would be
    '/outer/inner.' Do not use trailing slashes.

    Args:
        d (dict): the dictionary to search in
        path (str): the path string

    Returns:
        A tuple of
            (
                parent dictionary of entry,
                entry name,
                entry value (or None if not present),
            )
    """
    split = path.split("/")
    for name in split[:-1]:
        if not name:
            continue
        child = d.get(name, None)
        if child is None:
            d[name] = {}
        d = d[name]

    return d, split[-1], d.get(split[-1], None)


def _apply_changes(acq: dict, changes: list):
    """Applies all the changes in a list to an acquisition dict.
    All changes are performed in-place on the dict.

    Args:
        acq (dict): the dict to apply changes to
        changes (list): the list of changes
    """
    _LOGGER.debug(f"Applying {len(changes)} changes in conversion")
    for change in changes:
        if not isinstance(change, tuple):  # skip version tags
            continue

        op = change[0]
        args = change[1:]

        if op == "move":
            src_parent, src_name, src_value = _get_entry(acq, args[0])
            if src_value is not None:
                dst_parent, dst_name, _ = _get_entry(acq, args[1])
                dst_parent[dst_name] = src_value
                del src_parent[src_name]

        elif op == "copy":
            src_parent, src_name, src_value = _get_entry(acq, args[0])
            if src_value is not None:
                dst_parent, dst_name, _ = _get_entry(acq, args[1])
                dst_parent[dst_name] = copy.deepcopy(src_value)

        elif op == "del":
            parent, name, _ = _get_entry(acq, args[0])
            if name in parent:
                del parent[name]

        elif op == "set":
            parent, name, _ = _get_entry(acq, args[0])
            parent[name] = args[1]

        elif op == "modify":
            parent, name, value = _get_entry(acq, args[0])
            if value is not None:
                # Avoid sharing the same dicts between acquisitions
                if type(value) == dict:
                    value = copy.deepcopy(value)
                parent[name] = args[1](value)


def _get_changes_list(from_version: str, to_version: str):
    """Fetches a list of changes to convert from one version
    to another

    Args:
        from_version (str): the starting version number
        to_version (str): the ending version number

    Raises:
        ValueError if one of the versions is unsupported

    Returns:
        A list of changes to apply
    """
    try:
        from_version_index = _VERSION_TAGS.index(from_version)
        to_version_index = _VERSION_TAGS.index(to_version)
        if to_version_index >= from_version_index:
            changes = _UPGRADE_CHANGES
        else:
            changes = _DOWNGRADE_CHANGES

        start_index = changes.index(from_version)
        end_index = changes.index(to_version)

        return changes[start_index:end_index]
    except ValueError as e:
        _LOGGER.error(f"Got an invalid revision: {e}")
        raise ValueError(f"Unsupported version number in conversion: {e}")


# List of version tags in chronological order, used to distinguish upgrades/downgrades
_VERSION_TAGS = [
    "0.1",
    "0.2",
    "0.3",
]
LATEST_VERSION = _VERSION_TAGS[-1]
EARLIEST_VERSION = _VERSION_TAGS[0]


_UPGRADE_CHANGES = [  # Version tags appear after the operations needed to upgrade to that version
    "0.1",  # ==========================================================================
    ("move", "/info/board_info/model", "/settings/model"),
    ("move", "/info/board_info", "/info/readings"),
    ("move", "/info/readings/connection", "/info/connection"),
    ("move", "/info/parameters", "/settings/params"),
    ("move", "/info/analog_values", "/settings/registers/analog_registers"),
    ("move", "/info/control_registers", "/settings/registers/control_registers"),
    ("move", "/info/digital_registers", "/settings/registers/digital_registers"),
    ("move", "/trigger_settings/date", "/info/trigger_settings/date"),
    ("move", "/trigger_settings/levels", "/info/trigger_settings/levels"),
    ("move", "/trigger_settings/offset", "/info/trigger_settings/offset"),
    ("del", "/trigger_settings"),
    ("set", "file_version", "0.2"),
    "0.2",  # ==========================================================================
    ("del", "/model"),
    ("set", "file_version", "0.3"),
    "0.3",  # ==========================================================================
]
_DOWNGRADE_CHANGES = [  # Version tags appear after the operations needed to downgrade to that version
    "0.3",  # ==========================================================================
    ("copy", "/settings/model", "/model"),
    "0.2",  # ==========================================================================
    ("move", "/info/readings", "/info/board_info"),
    ("move", "/info/connection", "/info/board_info/connection"),
    ("move", "/settings/model", "/info/board_info/model"),
    ("move", "/settings/params", "/info/parameters"),
    ("move", "/settings/registers/analog_registers", "/info/analog_values"),
    ("move", "/settings/registers/control_registers", "/info/control_registers"),
    ("move", "/settings/registers/digital_registers", "/info/digital_registers"),
    ("del", "/settings"),
    ("copy", "/info/readout_settings/acq", "/trigger_settings/acq"),
    ("copy", "/info/readout_settings/readout_channels", "/trigger_settings/channels"),
    ("modify", "/trigger_settings/channels", (lambda x: ast.literal_eval(x))),
    ("move", "/info/trigger_settings/date", "/trigger_settings/date"),
    ("copy", "/info/readout_settings/lb", "/trigger_settings/lb"),
    ("copy", "/info/readout_settings/limit", "/trigger_settings/limit"),
    ("copy", "/info/readout_settings/ped", "/trigger_settings/ped"),
    ("copy", "/info/readout_settings/readoutEn", "/trigger_settings/readoutEn"),
    (
        "copy",
        "/info/readout_settings/readoutlookback",
        "/trigger_settings/readoutlookback",
    ),
    (
        "copy",
        "/info/readout_settings/readoutwindows",
        "/trigger_settings/readoutwindows",
    ),
    ("copy", "/info/readout_settings/singleEv", "/trigger_settings/singleEv"),
    ("copy", "/info/readout_settings/trig", "/trigger_settings/trig"),
    (
        "copy",
        "/info/readout_settings/writeaftertrig",
        "/trigger_settings/writeaftertrig",
    ),
    ("set", "file_version", "0.1"),
    "0.1",  # ==========================================================================
]
